########################################################################
# File name: test_e2e.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import unittest

import aioxmpp
import aioxmpp.private_xml as private_xml
import aioxmpp.xso as xso

from aioxmpp.utils import etree

from aioxmpp.e2etest import (
    blocking,
    blocking_timed,
    TestCase,
)


class TestPrivateXMLStorage(TestCase):

    @blocking
    @asyncio.coroutine
    def setUp(self):
        self.client, = yield from asyncio.gather(
            self.provisioner.get_connected_client(
                services=[
                    private_xml.PrivateXMLService
                ]
            ),
        )

    @blocking_timed
    @asyncio.coroutine
    def test_store_and_retrieve_xml_unregistered(self):
        tree = etree.fromstring(
            '<example xmlns="urn:example:unregistered">'
            '<payload xmlns="urn:example:unregistered">'
            'foobar'
            '</payload>'
            '</example>'
        )

        query = private_xml.xso.Query(None)
        query.unregistered_payload.append(tree)

        iq = aioxmpp.IQ(
            type_=aioxmpp.IQType.SET,
            payload=query,
        )

        yield from self.client.send(iq)

        query.unregistered_payload[0].clear()

        iq = aioxmpp.IQ(
            type_=aioxmpp.IQType.GET,
            payload=query,
        )

        retrieved = yield from self.client.send(iq)

        self.assertEqual(len(retrieved.unregistered_payload), 1)
        self.assertEqual(
            retrieved.unregistered_payload[0].tag,
            "{urn:example:unregistered}example"
        )
        self.assertEqual(len(retrieved.unregistered_payload[0]), 1)
        self.assertEqual(
            retrieved.unregistered_payload[0][0].tag,
            "{urn:example:unregistered}payload"
        )
        self.assertEqual(
            retrieved.unregistered_payload[0][0].text,
            "foobar"
        )

    @blocking_timed
    @asyncio.coroutine
    def test_store_and_retrieve_xml_registered(self):
        p = self.client.summon(private_xml.PrivateXMLService)

        @private_xml.Query.as_payload_class
        class Example(xso.XSO):
            TAG = ("urn:example:registered", "example")
            data = xso.Text(type_=xso.String())

            def __init__(self, text=""):
                self.data = text

        yield from p.set_private_xml(Example("foobar"))
        retrieved = yield from p.get_private_xml(Example())

        self.assertEqual(len(retrieved.unregistered_payload), 0)
        self.assertTrue(isinstance(retrieved.registered_payload, Example))
        self.assertEqual(retrieved.registered_payload.data, "foobar")



########################################################################
# File name: node.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.node` --- XMPP network nodes (clients, mostly)
#############################################################

This module contains functions to connect to an XMPP server, as well as
maintaining the stream. In addition, a client class which completely manages a
stream based on a presence setting is provided.

Using XMPP
==========

.. currentmodule:: aioxmpp

.. autoclass:: Client

.. autoclass:: PresenceManagedClient

.. currentmodule:: aioxmpp.node

.. class:: AbstractClient

   Alias of :class:`Client`.

   .. deprecated:: 0.8

      The alias will be removed in 1.0.

Connecting streams low-level
============================

.. autofunction:: discover_connectors

.. autofunction:: connect_xmlstream

Utilities
=========

.. autoclass:: UseConnected

"""
import asyncio
import contextlib
import logging
import warnings

from datetime import timedelta

import dns.resolver

import OpenSSL.SSL

import aiosasl

from . import (
    connector,
    network,
    protocol,
    errors,
    stream,
    callbacks,
    nonza,
    rfc3921,
    rfc6120,
    stanza,
    structs,
    security_layer,
    dispatcher,
    presence as mod_presence,
)
from .utils import namespaces


logger = logging.getLogger(__name__)


def lookup_addresses(loop, jid):
    addresses = yield from network.find_xmpp_host_addr(
        loop,
        jid.domain)

    return network.group_and_order_srv_records(addresses)


@asyncio.coroutine
def discover_connectors(
        domain: str,
        loop=None,
        logger=logger):
    """
    Discover all connection options for a domain, in descending order of
    preference.

    This coroutine returns options discovered from SRV records, or if none are
    found, the generic option using the domain name and the default XMPP client
    port.

    Each option is represented by a triple ``(host, port, connector)``.
    `connector` is a :class:`aioxmpp.connector.BaseConnector` instance which is
    suitable to connect to the given host and port.

    `logger` is the logger used by the function.

    The following sources are supported:

    * :rfc:`6120` SRV records. One option is returned per SRV record.

      If one of the SRV records points to the root name (``.``),
      :class:`ValueError` is raised (the domain specifically said that XMPP is
      not supported here).

    * :xep:`368` SRV records. One option is returned per SRV record.

    * :rfc:`6120` fallback process (only if no SRV records are found). One
      option is returned for the host name with the default XMPP client port.

    The options discovered from SRV records are mixed together, ordered by
    priority and then within priorities are shuffled according to their weight.
    Thus, if there are multiple records of equal priority, the result of the
    function is not deterministic.

    .. versionadded:: 0.6
    """

    domain_encoded = domain.encode("idna") + b"."
    starttls_srv_failed = False
    tls_srv_failed = False

    try:
        starttls_srv_records = yield from network.lookup_srv(
            domain_encoded,
            "xmpp-client",
        )
        starttls_srv_disabled = False
    except dns.resolver.NoNameservers as exc:
        starttls_srv_records = []
        starttls_srv_disabled = False
        starttls_srv_failed = True
        starttls_srv_exc = exc
        logger.debug("xmpp-client SRV lookup for domain %s failed "
                     "(may not be fatal)",
                     domain_encoded,
                     exc_info=True)
    except ValueError:
        starttls_srv_records = []
        starttls_srv_disabled = True

    try:
        tls_srv_records = yield from network.lookup_srv(
            domain_encoded,
            "xmpps-client",
        )
        tls_srv_disabled = False
    except dns.resolver.NoNameservers:
        tls_srv_records = []
        tls_srv_disabled = False
        tls_srv_failed = True
        logger.debug("xmpps-client SRV lookup for domain %s failed "
                     "(may not be fatal)",
                     domain_encoded,
                     exc_info=True)
    except ValueError:
        tls_srv_records = []
        tls_srv_disabled = True

    if starttls_srv_failed and (tls_srv_failed or tls_srv_records is None):
        # the failure is probably more useful as a diagnostic
        # if we find a good reason to allow this scenario, we might change it
        # later.
        raise starttls_srv_exc

    if starttls_srv_disabled and (tls_srv_disabled or tls_srv_records is None):
        raise ValueError(
            "XMPP not enabled on domain {!r}".format(domain),
        )

    if starttls_srv_records is None and tls_srv_records is None:
        # no SRV records published, fall back
        logger.debug(
            "no SRV records found for %s, falling back",
            domain,
        )
        return [
            (domain, 5222, connector.STARTTLSConnector()),
        ]

    starttls_srv_records = starttls_srv_records or []
    tls_srv_records = tls_srv_records or []

    srv_records = [
        (prio, weight, (host.decode("ascii"), port,
                        connector.STARTTLSConnector()))
        for prio, weight, (host, port) in starttls_srv_records
    ]

    srv_records.extend(
        (prio, weight, (host.decode("ascii"), port,
                        connector.XMPPOverTLSConnector()))
        for prio, weight, (host, port) in tls_srv_records
    )

    options = list(
        network.group_and_order_srv_records(srv_records)
    )

    logger.debug(
        "options for %s: %r",
        domain,
        options,
    )

    return options


@asyncio.coroutine
def _try_options(options, exceptions,
                 jid, metadata, negotiation_timeout, loop, logger):
    """
    Helper function for :func:`connect_xmlstream`.
    """
    for host, port, conn in options:
        logger.debug(
            "domain %s: trying to connect to %r:%s using %r",
            jid.domain, host, port, conn
        )
        try:
            transport, xmlstream, features = yield from conn.connect(
                loop,
                metadata,
                jid.domain,
                host,
                port,
                negotiation_timeout,
                base_logger=logger,
            )
        except OSError as exc:
            logger.warning(
                "connection failed: %s", exc
            )
            exceptions.append(exc)
            continue

        logger.debug(
            "domain %s: connection succeeded using %r",
            jid.domain,
            conn,
        )

        if not metadata.sasl_providers:
            return transport, xmlstream, features

        try:
            features = yield from security_layer.negotiate_sasl(
                transport,
                xmlstream,
                metadata.sasl_providers,
                negotiation_timeout=None,
                jid=jid,
                features=features,
            )
        except errors.SASLUnavailable as exc:
            protocol.send_stream_error_and_close(
                xmlstream,
                condition=errors.StreamErrorCondition.POLICY_VIOLATION,
                text=str(exc),
            )
            exceptions.append(exc)
            continue
        except Exception as exc:
            protocol.send_stream_error_and_close(
                xmlstream,
                condition=errors.StreamErrorCondition.UNDEFINED_CONDITION,
                text=str(exc),
            )
            raise

        return transport, xmlstream, features

    return None


@asyncio.coroutine
def connect_xmlstream(
        jid,
        metadata,
        negotiation_timeout=60.,
        override_peer=[],
        loop=None,
        logger=logger):
    """
    Prepare and connect a :class:`aioxmpp.protocol.XMLStream` to a server
    responsible for the given `jid` and authenticate against that server using
    the SASL mechansims described in `metadata`.

    :param jid: Address of the user for which the connection is made.
    :type jid: :class:`aioxmpp.JID`
    :param metadata: Connection metadata for configuring the TLS usage.
    :type metadata: :class:`~.security_layer.SecurityLayer`
    :param negotiation_timeout: Timeout for each individual negotiation step.
    :type negotiation_timeout: :class:`float` in seconds
    :param override_peer: Sequence of connection options which take precedence
                          over normal discovery methods.
    :type override_peer: sequence of (:class:`str`, :class:`int`,
                         :class:`~.BaseConnector`) triples
    :param loop: asyncio event loop to use (defaults to current)
    :type loop: :class:`asyncio.BaseEventLoop`
    :param logger: Logger to use (defaults to module-wide logger)
    :type logger: :class:`logging.Logger`
    :raises ValueError: if the domain from the `jid` announces that XMPP is not
                        supported at all.
    :raises aioxmpp.errors.TLSFailure: if all connection attempts fail and one
                                       of them is a :class:`~.TLSFailure`.
    :raises aioxmpp.errors.MultiOSError: if all connection attempts fail.
    :return: Transport, XML stream and the current stream features
    :rtype: tuple of (:class:`asyncio.BaseTransport`, :class:`~.XMLStream`,
            :class:`~.nonza.StreamFeatures`)

    The part of the `metadata` specifying the use of TLS is applied. If the
    security layer does not mandate TLS, the resulting XML stream may not be
    using TLS. TLS is used whenever possible.

    The connection options in `override_peer` are tried before any standardised
    discovery of connection options is made. Only if all of them fail,
    automatic discovery of connection options is performed.

    `loop` may be a :class:`asyncio.BaseEventLoop` to use. Defaults to the
    current event loop.

    If the domain from the `jid` announces that XMPP is not supported at all,
    :class:`ValueError` is raised. If no options are returned from
    :func:`discover_connectors` and `override_peer` is empty,
    :class:`ValueError` is raised, too.

    If all connection attempts fail, :class:`aioxmpp.errors.MultiOSError` is
    raised. The error contains one exception for each of the options discovered
    as well as the elements from `override_peer` in the order they were tried.

    A TLS problem is treated like any other connection problem and the other
    connection options are considered. However, if *all* connection options
    fail and the set of encountered errors includes a TLS error, the TLS error
    is re-raised instead of raising a :class:`aioxmpp.errors.MultiOSError`.

    Return a triple ``(transport, xmlstream, features)``. `transport`
    the underlying :class:`asyncio.Transport` which is used for the `xmlstream`
    :class:`~.protocol.XMLStream` instance. `features` is the
    :class:`aioxmpp.nonza.StreamFeatures` instance describing the features of
    the stream.

    .. versionadded:: 0.6

    .. versionchanged:: 0.8

       The explicit raising of TLS errors has been introduced. Before, TLS
       errors were treated like any other connection error, possibly masking
       configuration problems.
    """
    loop = asyncio.get_event_loop() if loop is None else loop

    options = list(override_peer)

    exceptions = []

    result = yield from _try_options(
        options,
        exceptions,
        jid, metadata, negotiation_timeout, loop, logger,
    )
    if result is not None:
        return result

    options = list((yield from discover_connectors(
        jid.domain,
        loop=loop,
        logger=logger,
    )))

    result = yield from _try_options(
        options,
        exceptions,
        jid, metadata, negotiation_timeout, loop, logger,
    )
    if result is not None:
        return result

    if not options and not override_peer:
        raise ValueError("no options to connect to XMPP domain {!r}".format(
            jid.domain
        ))

    for exc in exceptions:
        if isinstance(exc, errors.TLSFailure):
            raise exc

    raise errors.MultiOSError(
        "failed to connect to XMPP domain {!r}".format(jid.domain),
        exceptions
    )


class Client:
    """
    Base class to implement an XMPP client.

    :param local_jid: Jabber ID to connect as
    :type local_jid: :class:`~aioxmpp.JID`
    :param security_layer: Configuration for authentication and TLS
    :type security_layer: :class:`~aioxmpp.SecurityLayer`
    :param negotiation_timeout: Timeout for the individual stream negotiation
                                steps (bounds initial connection time)
    :type negotiation_timeout: :class:`datetime.timedelta`
    :param override_peer: Connection options which take precedence over the
                          standardised connection options
    :type override_peer: sequence of connection option triples
    :param max_inital_attempts: Maximum number of initial connection attempts
    :type max_initial_attempts: :class:`int`
    :param loop: Override the :mod:`asyncio` event loop to use
    :type loop: :class:`asyncio.BaseEventLoop` or :data:`None`
    :param logger: Logger to use instead of the default logger
    :type logger: :class:`logging.Logger` or :data:`None`

    These classes deal with managing the :class:`~aioxmpp.stream.StanzaStream`
    and the underlying :class:`~aioxmpp.protocol.XMLStream` instances. The
    abstract client provides functionality for connecting the xmlstream as well
    as signals which indicate changes in the stream state.

    The `security_layer` is best created using the
    :func:`aioxmpp.security_layer.make` function and must provide
    authentication for the given `local_jid`.

    If `loop` is given, it must be a :class:`asyncio.BaseEventLoop`
    instance. If it is not given, the current event loop is used.

    As a glue between the stanza stream and the XML stream, it also knows about
    stream management and performs stream management negotiation. It is
    specialized on client operations, which implies that it will try to keep
    the stream alive as long as wished by the client.

    The client will attempt to connect to the server(s) associated with the
    `local_jid`, using the prioritised `override_peer` setting or the
    standardised options for connecting (see :meth:`discover_connectors`). The
    initial connection attempt must succeed within `max_initial_attempts`.

    If the connection breaks after the first connection attempt, the client
    will try to resume the connection transparently. If the server supports
    stream management (:xep:`198`) with resumption, this is entirely
    transparent to all operations over the stream. If the stream is not
    resumable or the resumption fails and `allow_implicit_reconnect` is true,
    the application and services using the stream are notified about that. If,
    in that situation, `allow_implicit_reconnect` is false instead, the client
    stops with an error.

    The number of reconnection attempts is generally unbounded. The application
    is notified that the stream got interrupted with the
    :meth:`on_stream_suspended` is emitted. After reconnection,
    :meth:`on_stream_established` is emitted (possibly preceded by a
    :meth:`on_stream_destroyed` emission if the stream failed to resume). If
    the application wishes to bound the time the stream tries to transparently
    reconnect, it should connect to the :meth:`on_stream_suspended` signal and
    stop the stream as needed.

    The reconnection attempts are throttled using expenential backoff
    controlled by the :attr:`backoff_start`, :attr:`backoff_factor` and
    :attr:`backoff_cap` attributes.

    .. note::

       If `max_initial_attempts` is :data:`None`, the stream will try
       indefinitly to connect to the server even if the connection has
       never succeeded yet. This is may mask problems with the configuration of
       the client itself, because the client cannot successfully distinguish
       permanent problems arising from the configuration (of the client or the
       server) from problems arising from transient problems such as network
       faliures.

       This may severely degrade usabilty, because the client is then stuck in
       a connect loop without any usable feedback. Setting a bound for the
       initial connection attempt is usually better, for interactive
       applications an upper bound of 1 might make most sense (possibly the
       interactive application may retry on its own if the user did not
       indicate that they wish to do so after a timeout). We’ll leave the UX
       considerations up to you.

    .. versionchanged:: 0.4

       Since 0.4, support for legacy XMPP sessions has been implemented. Mainly
       for compatiblity with ejabberd.

    .. versionchanged:: 0.8

       The amount of initial connection attempts is now bounded by
       `max_initial_attempts`. The :meth:`on_stream_suspended` signal and the
       associated logic has been introduced.

    Controlling the client:

    .. automethod:: connected

    .. automethod:: start

    .. automethod:: stop

    .. autoattribute:: running

    .. attribute:: negotiation_timeout
        :annotation: = timedelta(seconds=60)

        The timeout applied to the connection process and the individual steps
        of negotiating the stream. See the `negotiation_timeout` argument to
        :func:`connect_xmlstream`.

    .. attribute:: override_peer

       A sequence of triples ``(host, port, connector)``, where `host` must be
       a host name or IP as string, `port` must be a port number and
       `connector` must be a :class:`aioxmpp.connector.BaseConnctor` instance.

       These connection options are passed to :meth:`connect_xmlstream` and
       thus take precedence over the options discovered using
       :meth:`discover_connectors`.

       .. note::

          If Stream Management is used and the peer server provided a location
          to connect to on resumption, that location is preferred even over the
          options set here.

       .. versionadded:: 0.6

    .. autoattribute:: resumption_timeout
        :annotation: = None

    Connection information:

    .. autoattribute:: established

    .. attribute:: established_event

        An :class:`asyncio.Event` which indicates that the stream is
        established. A stream is valid after resource binding and before it has
        been destroyed.

        While this event is cleared, :meth:`enqueue` fails with
        :class:`ConnectionError` and :meth:`send` blocks.

    .. autoattribute:: local_jid

    .. attribute:: stream

       The :class:`~aioxmpp.stream.StanzaStream` instance used by the node.

    .. attribute:: stream_features

       An instance of :class:`~aioxmpp.nonza.StreamFeatures`. This is the
       most-recently received stream features information (the one received
       right before resource binding).

       While no stream has been established yet, this is :data:`None`. During
       transparent re-negotiation, that information may be obsolete. However,
       when :attr:`before_stream_established` fires, the information is
       up-to-date.

    Sending stanzas:

    .. automethod:: send

    .. automethod:: enqueue

    Configuration of exponential backoff for reconnects:

    .. attribute:: backoff_start
       :annotation: = timedelta(1)

       When an underlying XML stream fails due to connectivity issues (generic
       :class:`OSError` raised), exponential backoff takes place before
       attempting to reconnect.

       The initial time to wait before reconnecting is described by
       :attr:`backoff_start`.

    .. attribute:: backoff_factor
       :annotation: = 1.2

       Each subsequent time a connection fails, the previous backoff time is
       multiplied with :attr:`backoff_factor`.

    .. attribute:: backoff_cap
       :annotation: = timedelta(60)

       The backoff time is capped to :attr:`backoff_cap`, to avoid having
       unrealistically high values.

    Signals:

    .. signal:: on_failure(err)

       This signal is fired when the client fails and stops.

    .. syncsignal:: before_stream_established()

       This coroutine signal is executed right before
       :meth:`on_stream_established` fires.

    .. signal:: on_stopped()

       Fires when the client stops gracefully. This is the counterpart to
       :meth:`on_failure`.

    .. signal:: on_stream_established()

       When the stream is established and resource binding took place, this
       event is fired. It means that the stream can now be used for XMPP
       interactions.

    .. signal:: on_stream_suspended(reason)

       The stream has been suspened due to a connection failure.

       :param reason: The exception which terminated the stream.
       :type reason: :class:`Exception`

       This signal may be immediately followed by a
       :meth:`on_stream_destroyed`, if the stream did not support stream
       resumption. Otherwise, a new connection is attempted transparently.

       In general, this signal exists solely for informational purposes. It
       can be used to drive a user interface which indicates that messages may
       be delivered with delay, because the underlying network is transiently
       interrupted.

       :meth:`on_stream_suspended` is not emitted if the stream was stopped on
       user request.

       Only :meth:`on_stream_destroyed` indicates that state was actually lost
       and that others most likely see or saw an unavailable presence broadcast
       for the resource.

       .. versionadded:: 0.8

    .. signal:: on_stream_destroyed(reason=None)

       This is called whenever a stream is destroyed. The conditions for this
       are the same as for
       :attr:`aioxmpp.stream.StanzaStream.on_stream_destroyed`.

       :param reason: An optional exception which indicates the reason for the
                      destruction of the stream.
       :type reason: :class:`Exception`

       This event can be used to know when to discard all state about the XMPP
       connection, such as roster information. Services implemented in
       :mod:`aioxmpp` generally subscribe to this signal to discard cached
       state.

       `reason` is optional. It is given if there is has been a specific
       exception which describes the cause for the stream destruction, such as
       a :class:`ConnectionError`.

       .. versionchanged:: 0.8

          The `reason` argument was added.

    Services:

    .. automethod:: summon

    Miscellaneous:

    .. attribute:: logger

       The :class:`logging.Logger` instance which is used by the
       :class:`Client`. This is the `logger` passed to the constructor or a
       logger derived from the fully qualified name of the class.

       .. versionadded:: 0.6

          The :attr:`logger` attribute was added.

    """

    on_failure = callbacks.Signal()
    on_stopped = callbacks.Signal()
    on_stream_destroyed = callbacks.Signal()
    on_stream_suspended = callbacks.Signal()
    on_stream_established = callbacks.Signal()

    before_stream_established = callbacks.SyncSignal()

    def __init__(self,
                 local_jid,
                 security_layer,
                 *,
                 negotiation_timeout=timedelta(seconds=60),
                 max_initial_attempts=4,
                 override_peer=[],
                 loop=None,
                 logger=None):
        super().__init__()
        self._local_jid = local_jid
        self._loop = loop or asyncio.get_event_loop()
        self._main_task = None
        self._security_layer = security_layer

        self._failure_future = asyncio.Future()
        self.logger = (logger or
                       logging.getLogger(".".join([
                           type(self).__module__,
                           type(self).__qualname__,
                       ])))

        self._backoff_time = None

        self._is_suspended = False

        # track whether the connection succeeded *at least once*
        # used to enforce max_initial_attempts
        self._had_connection = False
        self._nattempt = 0

        self._services = {}

        self.stream_features = None

        self.negotiation_timeout = negotiation_timeout
        self.backoff_start = timedelta(seconds=1)
        self.backoff_factor = 1.2
        self.backoff_cap = timedelta(seconds=60)
        self.override_peer = list(override_peer)
        self.established_event = asyncio.Event()
        self._max_initial_attempts = max_initial_attempts
        self._resumption_timeout = None

        self.on_stopped.logger = self.logger.getChild("on_stopped")
        self.on_failure.logger = self.logger.getChild("on_failure")
        self.on_stream_established.logger = \
            self.logger.getChild("on_stream_established")
        self.on_stream_destroyed.logger = \
            self.logger.getChild("on_stream_destroyed")
        self.on_stream_suspended.logger = \
            self.logger.getChild("on_stream_suspended")

        if logger is not None:
            stream_base_logger = self.logger
        else:
            stream_base_logger = logging.getLogger("aioxmpp")
        self.stream = stream.StanzaStream(
            local_jid.bare(),
            base_logger=stream_base_logger
        )

        self.stream._xxx_message_dispatcher = self.summon(
            dispatcher.SimpleMessageDispatcher,
        )

        self.stream._xxx_presence_dispatcher = self.summon(
            dispatcher.SimplePresenceDispatcher,
        )

        def send_warner(*args, **kwargs):
            warnings.warn("send() on StanzaStream is deprecated and will "
                          "be removed in 1.0. Use send() on the Client "
                          "instead.",
                          DeprecationWarning,
                          stacklevel=1)
            return self.send(*args, **kwargs)

        self.stream.send = send_warner

        def enqueue_warner(*args, **kwargs):
            warnings.warn("enqueue() on StanzaStream is deprecated and will "
                          "be removed in 1.0. Use enqueue() on the Client "
                          "instead.",
                          DeprecationWarning,
                          stacklevel=1)
            return self.enqueue(*args, **kwargs)

        self.stream.enqueue = enqueue_warner

    def _stream_failure(self, exc):
        if self._failure_future.done():
            self.logger.warning(
                "something is odd: failure future is already done ..."
            )
            return

        if not self._is_suspended:
            self.on_stream_suspended(exc)
            self._is_suspended = True

        self._failure_future.set_result(exc)
        self._failure_future = asyncio.Future()

    def _stream_destroyed(self, reason):
        if not self._is_suspended:
            if not isinstance(reason, stream.DestructionRequested):
                self.on_stream_suspended(reason)
            self._is_suspended = True

        if self.established_event.is_set():
            self.established_event.clear()
            self.on_stream_destroyed()

    def _on_main_done(self, task):
        try:
            task.result()
        except asyncio.CancelledError:
            # task terminated normally
            self.on_stopped()
        except Exception as err:
            self.logger.exception("main failed")
            self.on_failure(err)

    @asyncio.coroutine
    def _try_resume_stream_management(self, xmlstream, features):
        try:
            yield from self.stream.resume_sm(xmlstream)
        except errors.StreamNegotiationFailure as exc:
            self.logger.warn("failed to resume stream (%s)",
                             exc)
            return False
        return True

    @asyncio.coroutine
    def _negotiate_legacy_session(self):
        self.logger.debug(
            "remote server announces support for legacy sessions"
        )
        yield from self.stream._send_immediately(
            stanza.IQ(type_=structs.IQType.SET,
                      payload=rfc3921.Session())
        )
        self.logger.debug(
            "legacy session negotiated (upgrade your server!)"
        )

    @asyncio.coroutine
    def _negotiate_stream(self, xmlstream, features):
        server_can_do_sm = True
        try:
            features[nonza.StreamManagementFeature]
        except KeyError:
            if self.stream.sm_enabled:
                self.logger.warn("server isn’t advertising SM anymore")
                self.stream.stop_sm()
            server_can_do_sm = False

        self.logger.debug("negotiating stream (server_can_do_sm=%s)",
                          server_can_do_sm)

        if self.stream.sm_enabled:
            resumed = yield from self._try_resume_stream_management(
                xmlstream, features)
            if resumed:
                return features, resumed
        else:
            resumed = False

        self.stream_features = features
        self.stream.start(xmlstream)

        if not resumed:
            self.logger.debug("binding to resource")
            yield from self._bind()

        if server_can_do_sm:
            self.logger.debug("attempting to start stream management")
            try:
                yield from self.stream.start_sm(
                    resumption_timeout=self._resumption_timeout
                )
            except errors.StreamNegotiationFailure:
                self.logger.debug("stream management failed to start")
            self.logger.debug("stream management started")

        try:
            session_feature = features[rfc3921.SessionFeature]
        except KeyError:
            pass  # yay
        else:
            if not session_feature.optional:
                yield from self._negotiate_legacy_session()
            else:
                self.logger.debug(
                    "skipping optional legacy session negotiation"
                )

        self.established_event.set()

        yield from self.before_stream_established()

        self.on_stream_established()

        return features, resumed

    @asyncio.coroutine
    def _bind(self):
        iq = stanza.IQ(type_=structs.IQType.SET)
        iq.payload = rfc6120.Bind(resource=self._local_jid.resource)
        try:
            result = yield from self.stream._send_immediately(iq)
        except errors.XMPPError as exc:
            raise errors.StreamNegotiationFailure(
                "Resource binding failed: {}".format(exc)
            )

        self._local_jid = result.jid
        self.stream.local_jid = result.jid.bare()
        self.logger.info("bound to jid: %s", self._local_jid)

    @asyncio.coroutine
    def _main_impl(self):
        failure_future = self._failure_future

        override_peer = []
        if self.stream.sm_enabled:
            sm_location = self.stream.sm_location
            if sm_location:
                override_peer.append((
                    str(sm_location[0]),
                    sm_location[1],
                    connector.STARTTLSConnector(),
                ))
        override_peer += self.override_peer

        tls_transport, xmlstream, features = \
            yield from connect_xmlstream(
                self._local_jid,
                self._security_layer,
                negotiation_timeout=self.negotiation_timeout.total_seconds(),
                override_peer=override_peer,
                loop=self._loop,
                logger=self.logger)

        self._had_connection = True

        try:
            features, sm_resumed = yield from self._negotiate_stream(
                xmlstream,
                features)

            self._is_suspended = False
            self._backoff_time = None

            exc = yield from failure_future
            self.logger.error("stream failed: %s", exc)
            raise exc
        except asyncio.CancelledError:
            self.logger.info("client shutting down (on request)")
            # cancelled, this means a clean shutdown is requested
            yield from self.stream.close()
            raise
        finally:
            self.logger.info("stopping stream")
            self.stream.stop()

    @asyncio.coroutine
    def _main(self):
        with contextlib.ExitStack() as stack:
            stack.enter_context(
                self.stream.on_failure.context_connect(self._stream_failure)
            )
            stack.enter_context(
                self.stream.on_stream_destroyed.context_connect(
                    self._stream_destroyed)
            )
            while True:
                self._nattempt += 1
                self._failure_future = asyncio.Future()
                try:
                    yield from self._main_impl()
                except errors.StreamError as err:
                    if err.condition == errors.StreamErrorCondition.CONFLICT:
                        self.logger.debug("conflict!")
                        raise
                except (errors.StreamNegotiationFailure,
                        aiosasl.SASLError):
                    if self.stream.sm_enabled:
                        self.stream.stop_sm()
                    raise
                except (OSError, dns.resolver.NoNameservers,
                        OpenSSL.SSL.Error) as exc:
                    self.logger.info("connection error: (%s) %s",
                                     type(exc).__qualname__,
                                     exc)
                    if (not self._had_connection and
                            self._max_initial_attempts is not None and
                            self._nattempt >= self._max_initial_attempts):
                        self.logger.warning("out of connection attempts")
                        raise

                    if self._backoff_time is None:
                        self._backoff_time = self.backoff_start.total_seconds()
                    self.logger.debug("re-trying after %.1f seconds",
                                      self._backoff_time)
                    yield from asyncio.sleep(self._backoff_time)
                    self._backoff_time *= self.backoff_factor
                    if self._backoff_time > self.backoff_cap.total_seconds():
                        self._backoff_time = self.backoff_cap.total_seconds()
                    continue  # retry

    def start(self):
        """
        Start the client. If it is already :attr:`running`,
        :class:`RuntimeError` is raised.

        While the client is running, it will try to keep an XMPP connection
        open to the server associated with :attr:`local_jid`.
        """
        if self.running:
            raise RuntimeError("client already running")

        self._main_task = asyncio.ensure_future(
            self._main(),
            loop=self._loop
        )
        self._main_task.add_done_callback(self._on_main_done)

    def stop(self):
        """
        Stop the client. This sends a signal to the clients main task which
        makes it terminate.

        It may take some cycles through the event loop to stop the client
        task. To check whether the task has actually stopped, query
        :attr:`running`.
        """
        if not self.running:
            return

        self.logger.debug("stopping main task of %r", self, stack_info=True)
        self._main_task.cancel()

    # services

    def _summon(self, class_, visited):
        # this is essentially a topological sort algorithm
        try:
            return self._services[class_]
        except KeyError:
            if class_ in visited:
                raise ValueError("dependency loop")
            visited.add(class_)
            instance = class_(
                self,
                logger_base=self.logger,
                dependencies={
                    depclass: self._summon(depclass, visited)
                    for depclass in class_.PATCHED_ORDER_AFTER
                }
            )
            self._services[class_] = instance
            return instance

    def summon(self, class_):
        """
        Summon a :class:`~aioxmpp.service.Service` for the client.

        If the `class_` has already been summoned for the client, it’s instance
        is returned.

        Otherwise, all requirements for the class are first summoned (if they
        are not there already). Afterwards, the class itself is summoned and
        the instance is returned.
        """
        return self._summon(class_, set())

    # properties

    @property
    def local_jid(self):
        """
        The :class:`~aioxmpp.JID` the client currently has. While the
        client is disconnected, which parts of the :attr:`local_jid` can be
        relied upon depends on the authentication mechanism used. For example,
        using anonymous authentication, the server dictates even the local part
        of the JID and it will change after a reconnect. For more common
        authentication schemes (such as normal password-based authentication),
        the localpart is usually chosen by the client.

        For interoperability with different authentication schemes, code must
        invalidate all copies of this attribute when a
        :meth:`on_stream_established` or :meth:`on_stream_destroyed` event is
        emitted.

        Writing this attribute is not allowed, as changing the JID introduces a
        lot of issues with respect to reusability of the stream. Instanciate a
        new :class:`Client` if you need to change the bare part of the JID.

        .. note::

           Changing the resource between reconnects may be allowed later.
        """
        return self._local_jid

    @property
    def running(self):
        """
        true if the client is currently running, false otherwise.
        """
        return self._main_task is not None and not self._main_task.done()

    @property
    def established(self):
        """
        true if the stream is currently established (as defined in
        :attr:`on_stream_established`) and false otherwise.
        """
        return self.established_event.is_set()

    @property
    def resumption_timeout(self):
        """
        The maximum time as integer in seconds for which the server shall hold
        on to the session if the underlying transport breaks.

        This is only relevant if the server supports
        :xep:`Stream Management <198>` and the server may ignore the request
        for a maximum timeout and/or impose its own maximum. After the
        stream has been negotiated, :attr:`.StanzaStream.sm_max` holds the
        actual timeout announced by the server (may be :data:`None` if the
        server did not specify a timeout).

        The default value of :data:`None` does not request any specific
        timeout from the server and leaves it up to the server to decide.

        Setting a :attr:`resumption_timeout` of zero (0) disables resumption.

        .. versionadded:: 0.9
        """
        return self._resumption_timeout

    @resumption_timeout.setter
    def resumption_timeout(self, value):
        if (value is not None and
                (not isinstance(value, int) or isinstance(value, bool))):
            raise TypeError(
                "resumption_timeout must be int or None, got {!r}".format(
                    value
                )
            )
        if value is not None and value < 0:
            raise ValueError(
                "resumption timeout must be non-negative or None"
            )
        self._resumption_timeout = value

    def connected(self, *, presence=structs.PresenceState(False), **kwargs):
        """
        Return a :class:`.node.UseConnected` context manager which does not
        modify the presence settings.

        The keyword arguments are passed to the :class:`.node.UseConnected`
        context manager constructor.

        .. versionadded:: 0.8
        """
        return UseConnected(self, presence=presence, **kwargs)

    def enqueue(self, stanza, **kwargs):
        """
        Put a `stanza` in the internal transmission queue and return a token to
        track it.

        :param stanza: Stanza to send
        :type stanza: :class:`IQ`, :class:`Message` or :class:`Presence`
        :param kwargs: see :class:`StanzaToken`
        :raises ConnectionError: if the stream is not :attr:`established`
            yet.
        :return: token which tracks the stanza
        :rtype: :class:`StanzaToken`

        The `stanza` is enqueued in the active queue for transmission and will
        be sent on the next opportunity. The relative ordering of stanzas
        enqueued is always preserved.

        Return a fresh :class:`StanzaToken` instance which traks the progress
        of the transmission of the `stanza`. The `kwargs` are forwarded to the
        :class:`StanzaToken` constructor.

        This method calls :meth:`~.stanza.StanzaBase.autoset_id` on the stanza
        automatically.

        .. seealso::

           :meth:`send`
              for a more high-level way to send stanzas.

        .. versionchanged:: 0.10

            This method has been moved from
            :meth:`aioxmpp.stream.StanzaStream.enqueue`.
        """
        if not self.established_event.is_set():
            raise ConnectionError("stream is not ready")

        return self.stream._enqueue(stanza, **kwargs)

    @asyncio.coroutine
    def send(self, stanza, *, timeout=None, cb=None):
        """
        Send a stanza.

        :param stanza: Stanza to send
        :type stanza: :class:`~.IQ`, :class:`~.Presence` or :class:`~.Message`
        :param timeout: Maximum time in seconds to wait for an IQ response, or
                        :data:`None` to disable the timeout.
        :type timeout: :class:`~numbers.Real` or :data:`None`
        :param cb: Optional callback which is called synchronously when the
            reply is received (IQ requests only!)
        :raise OSError: if the underlying XML stream fails and stream
            management is not disabled.
        :raise aioxmpp.stream.DestructionRequested:
           if the stream is closed while sending the stanza or waiting for a
           response.
        :raise aioxmpp.errors.XMPPError: if an error IQ response is received
        :raise aioxmpp.errors.ErroneousStanza: if the IQ response could not be
            parsed
        :raise ValueError: if `cb` is given and `stanza` is not an IQ request.
        :return: IQ response :attr:`~.IQ.payload` or :data:`None`

        Send the stanza and wait for it to be sent. If the stanza is an IQ
        request, the response is awaited and the :attr:`~.IQ.payload` of the
        response is returned.

        If the stream is currently not ready, this method blocks until the
        stream is ready to send payload stanzas. Note that this may be before
        initial presence has been sent. To synchronise with that type of events,
        use the appropriate signals.

        The `timeout` as well as any of the exception cases referring to a
        "response" do not apply for IQ response stanzas, message stanzas or
        presence stanzas sent with this method, as this method only waits for
        a reply if an IQ *request* stanza is being sent.

        If `stanza` is an IQ request and the response is not received within
        `timeout` seconds, :class:`TimeoutError` (not
        :class:`asyncio.TimeoutError`!) is raised.

        If `cb` is given, `stanza` must be an IQ request (otherwise,
        :class:`ValueError` is raised before the stanza is sent). It must be a
        callable returning an awaitable. It receives the response stanza as
        first and only argument. The returned awaitable is awaited by
        :meth:`send` and the result is returned instead of the original
        payload. `cb` is called synchronously from the stream handling loop when
        the response is received, so it can benefit from the strong ordering
        guarantees given by XMPP XML Streams.

        The `cb` may also return :data:`None`, in which case :meth:`send` will
        simply return the IQ payload as if `cb` was not given. Since the return
        value of coroutine functions is awaitable, it is valid and supported to
        pass a coroutine function as `cb`.

        .. warning::

            Remember that it is an implementation detail of the event loop when
            a coroutine is scheduled after it awaited an awaitable; this implies
            that if the caller of :meth:`send` is merely awaiting the
            :meth:`send` coroutine, the strong ordering guarantees of XMPP XML
            Streams are lost.

            To regain those, use the `cb` argument.

        .. note::

            For the sake of readability, unless you really need the strong
            ordering guarantees, avoid the use of the `cb` argument. Avoid
            using a coroutine function unless you really need to.

        .. versionchanged:: 0.10

            * This method now waits until the stream is ready to send stanza¸
              payloads.
            * This method was moved from
              :meth:`aioxmpp.stream.StanzaStream.send`.

        .. versionchanged:: 0.9

            The `cb` argument was added.

        .. versionadded:: 0.8
        """
        if not self.running:
            raise ConnectionError("client is not running")

        if not self.established:
            self.logger.debug("send(%s): stream not established, waiting",
                              stanza)
            # wait for the stream to be established
            stopped_fut = self.on_stopped.future()
            failure_fut = self.on_failure.future()
            established_fut = asyncio.ensure_future(
                self.established_event.wait()
            )
            done, pending = yield from asyncio.wait(
                [
                    established_fut,
                    failure_fut,
                    stopped_fut,
                ],
                return_when=asyncio.FIRST_COMPLETED,
            )
            if not established_fut.done():
                established_fut.cancel()
            if failure_fut.done():
                if not stopped_fut.done():
                    stopped_fut.cancel()
                failure_fut.exception()
                raise ConnectionError("client failed to connect")
            if stopped_fut.done():
                raise ConnectionError("client shut down by user request")

            self.logger.debug("send(%s): stream established, sending")

        return (yield from self.stream._send_immediately(stanza,
                                                         timeout=timeout,
                                                         cb=cb))


class PresenceManagedClient(Client):
    """
    Client whose connection is controlled by its configured presence.

    .. seealso::

       :class:`Client`
          for a description of the arguments.

    The presence is set using :attr:`presence` or the :class:`PresenceServer`
    service. If the set presence is an *available* presence, the client is
    started (if it is not already running). If the set presence is an
    *unavailable* presence, the unavailable presence is broadcast and the
    client is stopped.

    While the start/stop interfaces of :class:`~.Client` are still available,
    using them may interfere with the behaviour of the presence automagic.

    The initial presence is set to `unavailable`, thus, the client will not
    connect immediately.

    .. autoattribute:: presence

    .. automethod:: set_presence

    .. automethod:: connected

    Signals:

    .. attribute:: on_presence_sent

       The event is fired after :meth:`.Client.on_stream_established` and after
       the current presence has been sent to the server as *initial presence*.

    .. versionchanged:: 0.8

       Since 0.8, the :class:`PresenceManagedClient` is implemented on top of
       :class:`PresenceServer`. Changing the presence via the
       :class:`PresenceServer` has the same effect as writing :attr:`presence`
       or calling :meth:`set_presence`.
    """

    on_presence_sent = callbacks.Signal()

    def __init__(self, jid, security_layer, **kwargs):
        super().__init__(jid, security_layer, **kwargs)
        self._presence_server = self.summon(mod_presence.PresenceServer)
        self._presence_server.on_presence_state_changed.connect(
            self._update_presence
        )
        self.on_stream_established.connect(self._handle_stream_established)

    def _handle_stream_established(self):
        self.on_presence_sent()

    def _update_presence(self):
        if self._presence_server.state.available:
            if not self.running:
                self.start()
        else:
            if self.running:
                self.stop()

    @property
    def presence(self):
        """
        Control or query the current presence state (see
        :class:`~.PresenceState`) of the client. Note that when
        reading, the property only returns the "set" value, not the actual
        value known to the server (and others). This may differ if the
        connection is still being established.

        .. seealso::

           Setting the presence state using :attr:`presence` clears the
           `status` of the presence. To set the status and state at once,
           use :meth:`set_presence`.

        Upon setting this attribute, the :class:`PresenceManagedClient` will do
        whatever neccessary to achieve the given presence. If the presence is
        an `available` presence, the client will attempt to connect to the
        server. If the presence is `unavailable` and the client is currently
        connected, it will disconnect.

        Instead of setting the presence to unavailable, :meth:`stop` can also
        be called. The :attr:`presence` attribute is *not* affected by calls to
        :meth:`start` or :meth:`stop`.
        """
        return self._presence_server.state

    @presence.setter
    def presence(self, value):
        call_update = value == self.presence
        self._presence_server.set_presence(value)
        if call_update:
            self._update_presence()

    def set_presence(self, state, status):
        """
        Set the presence `state` and `status` on the client. This has the same
        effects as writing `state` to :attr:`presence`, but the status of the
        presence is also set at the same time.

        `status` must be either a string or something which can be passed to
        :class:`dict`. If it is a string, the string is wrapped in a ``{None:
        status}`` dictionary. Otherwise, the dictionary is set as the
        :attr:`~.Presence.status` attribute of the presence stanza. It
        must map :class:`aioxmpp.structs.LanguageTag` instances to strings.

        The `status` is the text shown alongside the `state` (indicating
        availability such as *away*, *do not disturb* and *free to chat*).
        """
        self._presence_server.set_presence(state, status=status)

    def connected(self, **kwargs):
        """
        Return a :class:`.node.UseConnected` context manager which sets the
        presence to available.

        The keyword arguments are passed to the :class:`.node.UseConnected`
        context manager constructor.

        .. note::

           In contrast to the same method on :class:`Client`, this method
           implies setting an available presence.

        .. versionadded:: 0.6
        """
        return UseConnected(self, **kwargs)


class UseConnected:
    """
    Asynchronous context manager which connects and disconnects a
    :class:`.Client`.

    :param client: The client to manage
    :type client: :class:`.Client`
    :param timeout: Limit on the time it may take to start the client
    :type timeout: :class:`datetime.timedelta` or :data:`None`
    :param presence: Presence state to set on the client (deprecated)
    :type presence: :class:`.PresenceState`

    When the asynchronous context is entered (see :pep:`492`), the client is
    connected. This blocks until the client has finished connecting and the
    stream is established. If the client takes longer than `timeout` to
    connect, :class:`TimeoutError` is raised and the client is stopped. The
    context manager returns the :attr:`~.Client.stream` of the client.

    When the context is exited, the client is disconnected and the context
    manager waits for the client to cleanly shut down the stream.

    If the client is already connected when the context is entered, the
    connection is re-used and not shut down when the context is entered, but
    leaving the context still disconnects the client.

    If the `presence` refers to an available presence, the
    :class:`.PresenceServer` is :meth:`~.Client.summon`\ -ed on the `client`.
    The presence is set using :meth:`~.PresenceServer.set_presence` (clearing
    the :attr:`~.PresenceServer.status` and resetting
    :attr:`~.PresenceServer.priority` to 0) before the client is connected. If
    the client is already connected, the presence is set when the context is
    entered.

    .. deprecated:: 0.8

       The use of the `presence` argument is deprecated. The deprecation will
       happen in two phases:

       1. Until (but not including the release of) 1.0, passing a presence
          state which refers to an available presence will emit
          :class:`DeprecationWarning`. This *includes* the default of the
          argument, so unless an unavailable presence state is passed
          explicitly, all uses of :class:`UseConnected` emit that warning.

       2. Starting with 1.0, passing an available presence will raise
          :class:`ValueError`.

       3. Starting with a to-be-determined release after 1.0, passing the
          `presence` argument at all will raise :class:`TypeError`.

       Users which previously used the `presence` argument should use the
       :class:`.PresenceServer` service on the client and set the presence
       before using the context manager instead.

    .. autoattribute:: presence

       See the description of the `presence` argument.

       .. deprecated:: 0.8

          Using this attribute (for reading or writing) is deprecated and emits
          a deprecation warning.

    .. autoattribute:: timeout

       See the description of the `timeout` argument.

       .. deprecated:: 0.8

          Using this attribute (for reading or writing) is deprecated and emits
          a deprecation warning.
    """

    def __init__(self, client, *,
                 timeout=None,
                 presence=structs.PresenceState(True)):
        super().__init__()
        self._client = client
        self._timeout = timeout
        self._presence = presence
        if presence.available:
            warnings.warn(
                "using an available presence state for UseConnected is"
                " deprecated and will raise ValueError as of 1.0",
                DeprecationWarning,
                stacklevel=1,
            )

    @property
    def timeout(self):
        warnings.warn(
            "the timeout attribute is deprecated and will be removed in 1.0",
            DeprecationWarning,
            stacklevel=1,
        )
        return self._timeout

    @timeout.setter
    def timeout(self, value):
        warnings.warn(
            "the timeout attribute is deprecated and will be removed in 1.0",
            DeprecationWarning,
            stacklevel=1,
        )
        self._timeout = value

    @property
    def presence(self):
        warnings.warn(
            "the presence attribute is deprecated and will be removed in 1.0",
            DeprecationWarning,
            stacklevel=1,
        )
        return self._presence

    @presence.setter
    def presence(self, value):
        warnings.warn(
            "the presence attribute is deprecated and will be removed in 1.0",
            DeprecationWarning,
            stacklevel=1,
        )
        self._presence = value

    @asyncio.coroutine
    def __aenter__(self):
        if self._presence.available:
            svc = self._client.summon(
                mod_presence.PresenceServer
            )
            svc.set_presence(self._presence)

        if self._client.established:
            return self._client.stream

        conn_future = asyncio.Future()

        self._client.on_stream_established.connect(
            conn_future,
            self._client.on_stream_established.AUTO_FUTURE,
        )

        self._client.on_failure.connect(
            conn_future,
            self._client.on_failure.AUTO_FUTURE,
        )

        if not self._client.running:
            self._client.start()

        if self._timeout is not None:
            try:
                yield from asyncio.wait_for(
                    conn_future,
                    self._timeout.total_seconds(),
                )
            except asyncio.TimeoutError:
                self._client.stop()
                raise TimeoutError()
        else:
            yield from conn_future

        return self._client.stream

    @asyncio.coroutine
    def __aexit__(self, exc_type, exc_value, exc_traceback):
        if not self._client.running:
            return

        disconn_future = asyncio.Future()

        self._client.on_stopped.connect(
            disconn_future,
            self._client.on_stopped.AUTO_FUTURE,
        )

        self._client.on_failure.connect(
            disconn_future,
            self._client.on_failure.AUTO_FUTURE,
        )

        self._client.stop()

        try:
            yield from disconn_future
        except Exception:
            # we don’t want to re-raise that; the stream is dead, goal
            # achieved.
            pass



########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
from abc import abstractproperty

import aioxmpp.private_xml as private_xml
import aioxmpp.xso as xso


from aioxmpp.utils import namespaces


namespaces.xep0048 = "storage:bookmarks"


class Bookmark(xso.XSO):
    """
    A bookmark XSO abstract base class.

    Every XSO class registered as child of :class:`Storage` must be
    a :class:`Bookmark` subclass.

    Bookmarks must provide the following interface:

    .. autoattribute:: primary

    .. autoattribute:: secondary

    .. autoattribute:: name

    Equality is defined in terms of those properties:

    .. automethod:: __eq__

    It is highly recommended not to redefine :meth:`__eq__` in a
    subclass, if you do so make sure that the following axiom
    relating :meth:`__eq__`, :attr:`primary` and :attr:`secondary`
    holds::

        (type(a) == type(b) and
         a.primary == b.primary and
         a.secondary == b.secondary)

    if and only if::

        a == b

    Otherwise the generation of bookmark change signals is not
    guaranteed to be correct.
    """

    def __eq__(self, other):
        """
        Compare for equality by value and type.

        The value of a bookmark must be fully determined by the values
        of the :attr:`primary` and :attr:`secondary` properties.

        This is used for generating the bookmark list change signals
        and for the get-modify-set methods.
        """
        return (type(self) == type(other) and
                self.primary == other.primary and
                self.secondary == other.secondary)

    @abstractproperty
    def primary(self):
        """
        Return the primary category of the bookmark.

        The internal structure of the category is opaque to the code
        using it; only equality and hashing must be provided and
        operate by value. It is recommended that this be either a
        single datum (e.g. a string or JID) or a tuple of data items.

        Together with the type and :attr:`secondary` this must *fully*
        determine the value of the bookmark.

        This is used in the computation of the change
        signals. Bookmarks with different type or :attr:`primary`
        keys cannot be identified as changed from/to one another.
        """
        raise NotImplementedError  # pragma: no cover

    @abstractproperty
    def secondary(self):
        """
        Return the tuple of secondary categories of the bookmark.

        Together with the type and :attr:`primary` they must *fully*
        determine the value of the bookmark.

        This is used in the computation of the change signals. The
        categories in the tuple are ordered in decreasing precedence,
        when calculating which bookmarks have changed the ones which
        mismatch in the category with the lowest precedence are
        grouped together.

        The length of the tuple must be the same for all bookmarks of
        a type.
        """
        raise NotImplementedError  # pragma: no cover

    @abstractproperty
    def name(self):
        """
        The human-readable label or description of the bookmark.
        """
        raise NotImplementedError  # pragma: no cover


class Conference(Bookmark):
    """
    An bookmark for a groupchat.

    .. attribute:: name

       The name of the bookmark.

    .. attribute:: jid

       The jid under which the groupchat is accessible.

    .. attribute:: autojoin

       Whether to join automatically, when the client starts.

    .. attribute:: nick

       The nick to use in the groupchat.

    .. attribute:: password

       The password used to access the groupchat.
    """

    TAG = (namespaces.xep0048, "conference")

    autojoin = xso.Attr(tag="autojoin", type_=xso.Bool(), default=False)
    jid = xso.Attr(tag="jid", type_=xso.JID())
    name = xso.Attr(tag="name", type_=xso.String(), default=None)

    nick = xso.ChildText(
        (namespaces.xep0048, "nick"),
        default=None
    )
    password = xso.ChildText(
        (namespaces.xep0048, "password"),
        default=None
    )

    def __init__(self, name, jid, *, autojoin=False, nick=None, password=None):
        self.autojoin = autojoin
        self.jid = jid
        self.name = name
        self.nick = nick
        self.password = password

    def __repr__(self):
        return "Conference({!r}, {!r}, autojoin={!r}, " \
            "nick={!r}, password{!r})".\
            format(self.name, self.jid, self.autojoin, self.nick,
                   self.password)

    @property
    def primary(self):
        return self.jid

    @property
    def secondary(self):
        return (self.name, self.nick, self.password, self.autojoin)


class URL(Bookmark):
    """
    An URL bookmark.

    .. attribute:: name

       The name of the bookmark.

    .. attribute:: url

       The URL the bookmark saves.
    """
    TAG = (namespaces.xep0048, "url")

    name = xso.Attr(tag="name", type_=xso.String(), default=None)
    # XXX: we might want to use a URL type once we have one
    url = xso.Attr(tag="url", type_=xso.String())

    def __init__(self, name, url):
        self.name = name
        self.url = url

    def __repr__(self):
        return "URL({!r}, {!r})".format(self.name, self.url)

    @property
    def primary(self):
        return self.url

    @property
    def secondary(self):
        return (self.name,)


@private_xml.Query.as_payload_class
class Storage(xso.XSO):
    """
    The container for storing bookmarks.

    .. attribute:: bookmarks

       A :class:`~xso.XSOList` of bookmarks.
    """

    TAG = (namespaces.xep0048, "storage")

    bookmarks = xso.ChildList([URL, Conference])


def as_bookmark_class(xso_class):
    """
    Decorator to register `xso_class` as a custom bookmark class.

    This is necessary to store and retrieve such bookmarks.
    The registered class must be a subclass of the abstract base class
    :class:`Bookmark`.

    :raises TypeError: if `xso_class` is not a subclass of :class:`Bookmark`.
    """

    if not issubclass(xso_class, Bookmark):
        raise TypeError(
            "Classes registered as bookmark types must be Bookmark subclasses"
        )

    Storage.register_child(
        Storage.bookmarks,
        xso_class
    )

    return xso_class



########################################################################
# File name: testutils.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
This module contains utilities used for testing aioxmpp code. These
utilities themselves are tested, which is meta, but cool.
"""
import asyncio
import collections
import contextlib
import functools
import logging
import time
import unittest
import unittest.mock

from datetime import timedelta

import aioxmpp.callbacks as callbacks
import aioxmpp.xso as xso
import aioxmpp.nonza as nonza

from aioxmpp.utils import etree


GLOBAL_TIMEOUT_FACTOR = 1.0

_monotonic_info = time.get_clock_info("monotonic")
GLOBAL_TIMEOUT_FACTOR *= max(_monotonic_info.resolution, 0.0015) / 0.0015


logging.getLogger(__name__).debug("using GLOBAL_TIMEOUT_FACTOR = %.3f",
                                  GLOBAL_TIMEOUT_FACTOR)


def get_timeout(base):
    return base * GLOBAL_TIMEOUT_FACTOR


# FIXME: find a way to detect Travis CI and use 2.0 there, 1.0 otherwise.
# 1.0 is sufficient normally, but on Travis we sometimes get spurious failures.
DEFAULT_TIMEOUT = get_timeout(2.0)


def make_protocol_mock():
    return unittest.mock.Mock([
        "connection_made",
        "eof_received",
        "connection_lost",
        "data_received",
        "pause_writing",
        "resume_writing",
    ])


def run_coroutine(coroutine, timeout=DEFAULT_TIMEOUT, loop=None):
    if not loop:
        loop = asyncio.get_event_loop()
    return loop.run_until_complete(
        asyncio.wait_for(
            coroutine,
            timeout=timeout))


def run_coroutine_with_peer(
        coroutine,
        peer_coroutine,
        timeout=1.0,
        loop=None):
    loop = loop or asyncio.get_event_loop()

    local_future = asyncio.ensure_future(coroutine, loop=loop)
    remote_future = asyncio.ensure_future(peer_coroutine, loop=loop)

    done, pending = loop.run_until_complete(
        asyncio.wait(
            [
                local_future,
                remote_future,
            ],
            timeout=timeout,
            return_when=asyncio.FIRST_EXCEPTION)
    )
    if not done:
        raise asyncio.TimeoutError("Test timed out")

    if pending:
        pending_fut = next(iter(pending))
        pending_fut.cancel()
        fut = next(iter(done))
        try:
            fut.result()
        except:
            # everything is fine, the other one failed
            raise
        else:
            if pending_fut == remote_future:
                raise asyncio.TimeoutError(
                    "Peer coroutine did not return in time")
            else:
                raise asyncio.TimeoutError(
                    "Coroutine under test did not return in time")

    if local_future.exception():
        # re-throw the error properly
        local_future.result()

    remote_future.result()
    return local_future.result()


def make_listener(instance):
    """
    Return a :class:`unittest.mock.Mock` which has children connected to each
    :class:`aioxmpp.callbacks.Signal` of `instance`.

    The children are named exactly like the signals.
    """
    result = unittest.mock.Mock([])
    names = {
        name
        for type_ in type(instance).__mro__
        for name in type_.__dict__
    }
    for name in names:
        signal = getattr(instance, name)
        if not isinstance(signal, callbacks.AdHocSignal):
            continue
        cb = unittest.mock.Mock()
        setattr(result, name, cb)
        cb.return_value = None
        signal.connect(cb)
    return result


class FilterMock(unittest.mock.Mock):
    def __init__(self):
        super().__init__([
            "register",
            "unregister",
            "filter",
        ])
        self.context_register = unittest.mock.MagicMock([
            "__enter__",
            "__exit__",
        ])


class ConnectedClientMock(unittest.mock.Mock):
    on_stream_established = callbacks.Signal()
    on_stream_destroyed = callbacks.Signal()
    on_failure = callbacks.Signal()
    on_stopped = callbacks.Signal()

    before_stream_established = callbacks.SyncSignal()

    negotiation_timeout = timedelta(milliseconds=100)

    def __init__(self):
        super().__init__([
            "stream",
            "start",
            "stop",
            "set_presence",
            "local_jid",
            "enqueue",
        ])

        self.established = True

        self.stream_features = nonza.StreamFeatures()
        self.stream.on_message_received = callbacks.AdHocSignal()
        self.stream.on_presence_received = callbacks.AdHocSignal()
        self.stream.on_stream_destroyed = callbacks.AdHocSignal()
        self.stream.app_inbound_message_filter = FilterMock()
        self.stream.app_inbound_presence_filter = FilterMock()
        self.stream.app_outbound_message_filter = FilterMock()
        self.stream.app_outbound_presence_filter = FilterMock()
        self.stream.service_inbound_message_filter = FilterMock()
        self.stream.service_inbound_presence_filter = FilterMock()
        self.stream.service_outbound_message_filter = FilterMock()
        self.stream.service_outbound_presence_filter = FilterMock()
        self.stream.on_stream_destroyed = callbacks.AdHocSignal()
        self.stream.send_iq_and_wait_for_reply.side_effect = \
            AssertionError("use of deprecated function")
        self.stream.send.side_effect = \
            AssertionError("use of deprecated function")
        self.stream.enqueue.side_effect = \
            AssertionError("use of deprecated function")
        self.send = CoroutineMock()
        self.stream.enqueue_stanza = self.stream.enqueue
        self.mock_services = {}

    def _get_child_mock(self, **kw):
        return unittest.mock.Mock(**kw)

    def summon(self, cls):
        try:
            return self.mock_services[cls]
        except KeyError:
            raise AssertionError("service class not provisioned in mock")


def make_connected_client():
    return ConnectedClientMock()


class CoroutineMock(unittest.mock.Mock):
    delay = 0

    @asyncio.coroutine
    def __call__(self, *args, **kwargs):
        result = super().__call__(*args, **kwargs)
        yield from asyncio.sleep(self.delay)
        return result


class SSLWrapperMock:
    """
    Mock for :class:`aioxmpp.ssl_wrapper.STARTTLSableTransportProtocol`.

    The *protocol* must be an :class:`XMLStreamMock`, as the
    :class:`SSLWrapperMock` depends on some private attributes to ensure the
    sequence of events is correct.
    """

    # FIXME: this mock is not covered by tests :(

    def __init__(self, loop, protocol):
        super().__init__()
        self._loop = loop
        self._protocol = protocol

    @asyncio.coroutine
    def starttls(self, ssl_context=None, post_handshake_callback=None):
        """
        Override the STARTTLS sequence. Instead of actually starting a TLS
        transport on the existing socket, only make sure that the test expects
        starttls to happen now. If so, return fake information on the TLS
        transport.
        """

        tester = self._protocol._tester
        tester.assertFalse(self._protocol._closed)
        tester.assertTrue(self._protocol._action_sequence,
                          "Unexpected client action (no actions left)")
        to_recv, to_send = self._protocol._action_sequence.pop(0)
        tester.assertTrue(to_recv.startswith("!starttls"),
                          "Unexpected starttls attempt by the client")
        return self, None

    def close(self):
        pass


class InteractivityMock:
    def __init__(self, tester, *, loop=None):
        super().__init__()
        self._loop = loop or asyncio.get_event_loop()
        self._tester = tester

    def _check_done(self):
        if not self._done.done() and not self._actions:
            self._done.set_result(None)

    def _pop_and_call_and_catch(self, fun, *args):
        @functools.wraps(fun)
        def wrap():
            try:
                self._actions.pop(0)
                fun(*args)
            except Exception as err:
                self._done.set_exception(err)
            else:
                self._check_done()
        self._loop.call_soon(wrap)

    def _format_unexpected_action(self, action_name, reason):
        return "unexpected {name} ({reason})".format(
            name=action_name,
            reason=reason
        )

    def _basic(self, name, action_cls):
        self._tester.assertTrue(
            self._actions,
            self._format_unexpected_action(name, "no actions left"),
        )
        head = self._actions[0]
        self._tester.assertIsInstance(
            head, action_cls,
            self._format_unexpected_action(name, "expected something else"),
        )
        self._actions.pop(0)
        self._execute_response(head.response)

    def _execute_response(self, response):
        if response is None:
            return

        try:
            do = response.do
        except AttributeError:
            # we have the for loop outside this except: block, to have a
            # clearer traceback.
            if not hasattr(response, "__iter__"):
                raise RuntimeError("test specification incorrect: "
                                   "unknown response type: "+repr(response))
        else:
            self._execute_single(do)
            return

        for item in response:
            self._execute_response(item)


_Write = collections.namedtuple("Write", ["data", "response"])
_STARTTLS = collections.namedtuple("STARTTLS",
                                   ["ssl_context",
                                    "post_handshake_callback",
                                    "response"])
GenericTransportAction = collections.namedtuple(
    "GenericTransportAction",
    ["response"])
_LoseConnection = collections.namedtuple("LoseConnection", ["exc"])


class TransportMock(InteractivityMock,
                    asyncio.ReadTransport,
                    asyncio.WriteTransport):
    class Write(_Write):
        def __new__(cls, data, *, response=None):
            return _Write.__new__(cls, data=data, response=response)
        replace = _Write._replace

    class STARTTLS(_STARTTLS):
        def __new__(cls, ssl_context, post_handshake_callback, *,
                    response=None):
            return _STARTTLS.__new__(cls,
                                     ssl_context,
                                     post_handshake_callback,
                                     response=response)
        replace = _STARTTLS._replace

    class Abort(GenericTransportAction):
        def __new__(cls, *, response=None):
            return GenericTransportAction.__new__(cls, response=response)
        replace = GenericTransportAction._replace

    class WriteEof(GenericTransportAction):
        def __new__(cls, *, response=None):
            return GenericTransportAction.__new__(cls, response=response)
        replace = GenericTransportAction._replace

    class Receive(collections.namedtuple("Receive", ["data"])):
        def do(self, transport, protocol):
            protocol.data_received(self.data)

    class Close(GenericTransportAction):
        def __new__(cls, *, response=None):
            return GenericTransportAction.__new__(cls, response=response)
        replace = GenericTransportAction._replace

    class ReceiveEof:
        def __repr__(self):
            return "ReceiveEof()"

        def do(self, transport, protocol):
            protocol.eof_received()

    class MakeConnection:
        def __repr__(self):
            return "MakeConnection()"

        def do(self, transport, protocol):
            transport._connection_made = True
            protocol.connection_made(transport)

    class LoseConnection(_LoseConnection):
        def __new__(cls, exc=None):
            return _LoseConnection.__new__(cls, exc)

        def do(self, transport, protocol):
            protocol.connection_lost(self.exc)
            transport._connection_made = False

    def __init__(self, tester, protocol, *, with_starttls=False, loop=None):
        super().__init__(tester, loop=loop)
        self._protocol = protocol
        self._actions = None
        self._connection_made = False
        self._rxd = []
        self._queue = asyncio.Queue()
        self._with_starttls = with_starttls

    def _previously(self):
        buf = b"".join(self._rxd)
        result = [" (previously: "]
        if len(buf) > 100:
            result.append("[ {} more bytes ]".format(len(buf) - 100))
            buf = buf[-100:]
        result.append(str(buf)[1:])
        result.append(")")
        return "".join(result)

    def _format_unexpected_action(self, action_name, reason):
        return (
            super()._format_unexpected_action(action_name, reason) +
            self._previously()
        )

    def _execute_single(self, do):
        do(self, self._protocol)

    @asyncio.coroutine
    def run_test(self, actions, stimulus=None, partial=False):
        self._done = asyncio.Future()
        self._actions = actions
        if not self._connection_made:
            self._execute_response(self.MakeConnection())
        if stimulus:
            if isinstance(stimulus, bytes):
                self._execute_response(self.Receive(stimulus))
            else:
                self._execute_response(stimulus)

        while not self._queue.empty() or self._actions:
            done, pending = yield from asyncio.wait(
                [
                    self._queue.get(),
                    self._done
                ],
                return_when=asyncio.FIRST_COMPLETED
            )
            if self._done not in pending:
                # raise if error
                self._done.result()
                done.remove(self._done)

            if done:
                value_future = next(iter(done))
                action, *args = value_future.result()
                if action == "write":
                    yield from self._write(*args)
                elif action == "write_eof":
                    yield from self._write_eof(*args)
                elif action == "close":
                    yield from self._close(*args)
                elif action == "abort":
                    yield from self._abort(*args)
                elif action == "starttls":
                    yield from self._starttls(*args)
                else:
                    assert False

            if self._done not in pending:
                break

        if self._connection_made and not partial:
            self._execute_response(self.LoseConnection())

    def can_write_eof(self):
        return True

    @asyncio.coroutine
    def _write_eof(self):
        self._basic("write_eof", self.WriteEof)

    @asyncio.coroutine
    def _write(self, data):
        self._tester.assertTrue(
            self._actions,
            "unexpected write (no actions left)"+self._previously()
        )
        head = self._actions[0]
        self._tester.assertIsInstance(head, self.Write)
        expected_data = head.data
        if not expected_data.startswith(data):
            logging.info("expected: %r", expected_data)
            logging.info("got this: %r", data)
            self._tester.assertEqual(
                expected_data[:len(data)],
                bytes(data),
                "mismatch of expected and written data"+self._previously()
            )
        self._rxd.append(data)
        expected_data = expected_data[len(data):]
        if not expected_data:
            self._actions.pop(0)
            self._execute_response(head.response)
        else:
            self._actions[0] = head.replace(data=expected_data)

    @asyncio.coroutine
    def _abort(self):
        self._basic("abort", self.Abort)

    @asyncio.coroutine
    def _close(self):
        self._basic("close", self.Close)

    @asyncio.coroutine
    def _starttls(self, ssl_context, post_handshake_callback, fut):
        self._tester.assertTrue(
            self._actions,
            self._format_unexpected_action("starttls", "no actions left"),
        )
        head = self._actions[0]
        self._tester.assertIsInstance(
            head, self.STARTTLS,
            self._format_unexpected_action("starttls",
                                           "expected something else"),
        )
        self._actions.pop(0)

        self._tester.assertEqual(
            ssl_context,
            head.ssl_context,
            "mismatched starttls argument")
        self._tester.assertEqual(
            post_handshake_callback,
            head.post_handshake_callback,
            "mismatched starttls argument")

        if post_handshake_callback:
            try:
                yield from post_handshake_callback(self)
            except Exception as exc:
                fut.set_exception(exc)
            else:
                fut.set_result(None)
        else:
            fut.set_result(None)

        self._execute_response(head.response)

    def write(self, data):
        self._queue.put_nowait(("write", data))

    def write_eof(self):
        self._queue.put_nowait(("write_eof", ))

    def abort(self):
        self._queue.put_nowait(("abort", ))

    def close(self):
        self._queue.put_nowait(("close", ))

    def can_starttls(self):
        return self._with_starttls

    @asyncio.coroutine
    def starttls(self, ssl_context=None, post_handshake_callback=None):
        if not self._with_starttls:
            raise RuntimeError("STARTTLS not supported")

        fut = asyncio.Future()
        self._queue.put_nowait(
            ("starttls", ssl_context, post_handshake_callback, fut)
        )
        yield from fut


class XMLStreamMock(InteractivityMock):
    class Receive(collections.namedtuple("Receive", ["obj"])):
        def do(self, xmlstream):
            clsmap = xmlstream.stanza_parser.get_class_map()
            cls = type(self.obj)
            xmlstream._tester.assertIn(
                cls, clsmap,
                "no handler registered for {}".format(cls)
            )
            clsmap[cls](self.obj)

    class Fail(collections.namedtuple("Fail", ["exc"])):
        def do(self, xmlstream):
            xmlstream._exception = self.exc
            for fut in xmlstream._error_futures:
                if not fut.done():
                    fut.set_exception(self.exc)
            xmlstream.on_closing(self.exc)

    class Send(collections.namedtuple("Send", ["obj", "response"])):
        def __new__(cls, obj, *, response=None):
            return super().__new__(cls, obj, response)

    class Reset(collections.namedtuple("Reset", ["response"])):
        def __new__(cls, *, response=None):
            return super().__new__(cls, response)

    class Close(collections.namedtuple("Close", ["response"])):
        def __new__(cls, *, response=None):
            return super().__new__(cls, response)

    class Abort(collections.namedtuple("Abort", ["response"])):
        def __new__(cls, *, response=None):
            return super().__new__(cls, response)

    class Mute(collections.namedtuple("Mute", ["response"])):
        def __new__(cls, *, response=None):
            return super().__new__(cls, response)

    class Unmute(collections.namedtuple("Unmute", ["response"])):
        def __new__(cls, *, response=None):
            return super().__new__(cls, response)

    class STARTTLS(collections.namedtuple("STARTTLS", [
            "ssl_context", "post_handshake_callback", "response"])):
        def __new__(cls, ssl_context, post_handshake_callback,
                    *, response=None):
            return super().__new__(cls,
                                   ssl_context,
                                   post_handshake_callback,
                                   response)

    on_closing = callbacks.Signal()
    on_deadtime_soft_limit_tripped = callbacks.Signal()

    def __init__(self, tester, *, loop=None):
        super().__init__(tester, loop=loop)
        self._queue = asyncio.Queue()
        self._exception = None
        self._closed = False
        self.stanza_parser = xso.XSOParser()
        self.can_starttls_value = False
        self._error_futures = []

    def _execute_single(self, do):
        do(self)

    @asyncio.coroutine
    def run_test(self, actions,
                 stimulus=None):
        self._done = asyncio.Future()
        self._actions = actions

        self._execute_response(stimulus)

        while not self._queue.empty() or self._actions:
            done, pending = yield from asyncio.wait(
                [
                    self._queue.get(),
                    self._done
                ],
                return_when=asyncio.FIRST_COMPLETED
            )

            if self._done not in pending:
                # raise if error
                self._done.result()
                done.remove(self._done)

            if done:
                value_future = next(iter(done))
                action, *args = value_future.result()
                if action == "send":
                    yield from self._send_xso(*args)
                elif action == "reset":
                    yield from self._reset(*args)
                elif action == "close":
                    yield from self._close(*args)
                elif action == "starttls":
                    yield from self._starttls(*args)
                elif action == "abort":
                    yield from self._abort(*args)
                elif action == "mute":
                    yield from self._mute(*args)
                elif action == "unmute":
                    yield from self._unmute(*args)
                else:
                    assert False

            if self._done not in pending:
                break

    @asyncio.coroutine
    def _send_xso(self, obj):
        self._tester.assertTrue(
            self._actions,
            self._format_unexpected_action(
                "send_xso("+repr(obj)+")",
                "no actions left")
        )
        head = self._actions[0]
        self._tester.assertIsInstance(
            head, self.Send,
            self._format_unexpected_action(
                "send_xso",
                "expected something different")
        )

        t1 = etree.Element("root")
        obj.unparse_to_node(t1)
        t2 = etree.Element("root")
        head.obj.unparse_to_node(t2)

        self._tester.assertSubtreeEqual(t1, t2)
        self._actions.pop(0)
        self._execute_response(head.response)

    @asyncio.coroutine
    def _reset(self):
        self._basic("reset", self.Reset)

    @asyncio.coroutine
    def _mute(self):
        self._basic("mute", self.Mute)

    @asyncio.coroutine
    def _unmute(self):
        self._basic("unmute", self.Unmute)

    @asyncio.coroutine
    def _abort(self):
        self._basic("abort", self.Abort)
        self._exception = ConnectionError("not connected")
        for fut in self._error_futures:
            if not fut.done():
                fut.set_exception(self._exception)

    @asyncio.coroutine
    def _close(self):
        self._basic("close", self.Close)
        self._exception = ConnectionError("not connected")
        self.on_closing(None)
        for fut in self._error_futures:
            if not fut.done():
                fut.set_exception(self._exception)

    @asyncio.coroutine
    def _starttls(self, ssl_context, post_handshake_callback, fut):
        self._tester.assertTrue(
            self._actions,
            self._format_unexpected_action("starttls", "no actions left"),
        )
        head = self._actions[0]
        self._tester.assertIsInstance(
            head, self.STARTTLS,
            self._format_unexpected_action("starttls",
                                           "expected something else"),
        )
        self._actions.pop(0)

        self._tester.assertEqual(
            ssl_context,
            head.ssl_context,
            "mismatched starttls argument")
        self._tester.assertEqual(
            post_handshake_callback,
            head.post_handshake_callback,
            "mismatched starttls argument")

        if post_handshake_callback:
            try:
                yield from post_handshake_callback(self.transport)
            except Exception as exc:
                fut.set_exception(exc)
            else:
                fut.set_result(None)
        else:
            fut.set_result(None)

        self._execute_response(head.response)

    def send_xso(self, obj):
        if self._exception:
            raise self._exception
        self._queue.put_nowait(("send", obj))

    def reset(self):
        if self._exception:
            raise self._exception
        self._queue.put_nowait(("reset",))

    def abort(self):
        if self._exception:
            raise self._exception
        self._queue.put_nowait(("abort",))

    def close(self):
        if self._exception:
            raise self._exception
        self._queue.put_nowait(("close",))

    @asyncio.coroutine
    def starttls(self, ssl_context, post_handshake_callback=None):
        if self._exception:
            raise self._exception

        fut = asyncio.Future()
        self._queue.put_nowait(
            ("starttls", ssl_context, post_handshake_callback, fut)
        )
        yield from fut

    @asyncio.coroutine
    def close_and_wait(self):
        fut = asyncio.Future()
        self.on_closing.connect(fut, self.on_closing.AUTO_FUTURE)
        self.close()
        try:
            yield from fut
        except Exception:
            pass

    @contextlib.contextmanager
    def mute(self):
        self._queue.put_nowait(("mute",))
        try:
            yield
        finally:
            self._queue.put_nowait(("unmute",))

    def can_starttls(self):
        return self.can_starttls_value

    def error_future(self):
        fut = asyncio.Future()
        self._error_futures.append(fut)
        return fut


if not hasattr(unittest.mock.Mock, "assert_not_called"):
    def _assert_not_called(m):
        if any(not call[0] for call in m.mock_calls):
            raise AssertionError(
                "expected {!r} to not have been called. "
                "Called {} times".format(
                    m._mock_name or 'mock',
                    m.call_count)
            )
    unittest.mock.Mock.assert_not_called = _assert_not_called
    del _assert_not_called



########################################################################
# File name: sasl.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.sasl` -- SASL helpers
####################################

This module is used to implement SASL in :mod:`aioxmpp.security_layer`. It
provides a state machine for use by the different SASL mechanisms and
implementations of some SASL mechansims.

It provides an XMPP adaptor for :mod:`aiosasl`.

.. autoclass:: SASLXMPPInterface

The XSOs for SASL authentication can be found in :mod:`aioxmpp.nonza`.

"""

import asyncio
import logging

import aiosasl

from . import protocol, nonza

logger = logging.getLogger(__name__)


class SASLXMPPInterface(aiosasl.SASLInterface):
    def __init__(self, xmlstream):
        super().__init__()
        self.xmlstream = xmlstream
        self.timeout = None

    @asyncio.coroutine
    def _send_sasl_node_and_wait_for(self, node):
        node = yield from protocol.send_and_wait_for(
            self.xmlstream,
            [node],
            [
                nonza.SASLChallenge,
                nonza.SASLFailure,
                nonza.SASLSuccess
            ],
            timeout=self.timeout
        )

        state = node.TAG[1]

        if state == "failure":
            xmpp_error = node.condition[1]
            text = node.text
            raise aiosasl.SASLFailure(xmpp_error, text=text)

        if hasattr(node, "payload"):
            payload = node.payload
        else:
            payload = None

        return state, payload

    @asyncio.coroutine
    def initiate(self, mechanism, payload=None):
        with self.xmlstream.mute():
            return (yield from self._send_sasl_node_and_wait_for(
                nonza.SASLAuth(mechanism=mechanism,
                               payload=payload)))

    @asyncio.coroutine
    def respond(self, payload):
        with self.xmlstream.mute():
            return (yield from self._send_sasl_node_and_wait_for(
                nonza.SASLResponse(payload=payload)
            ))

    @asyncio.coroutine
    def abort(self):
        try:
            next_state, payload = yield from self._send_sasl_node_and_wait_for(
                nonza.SASLAbort()
            )
        except aiosasl.SASLFailure as err:
            self._state = "failure"
            if err.opaque_error != "aborted":
                raise
            return "failure", None
        else:
            raise aiosasl.SASLFailure(
                "aborted",
                text="unexpected non-failure after abort: "
                "{}".format(self._state)
            )



########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest
import unittest.mock

import aioxmpp
import aioxmpp.disco
import aioxmpp.version.service as version_svc
import aioxmpp.version.xso as version_xso
import aioxmpp.service

from aioxmpp.utils import namespaces

from aioxmpp.testutils import (
    make_connected_client,
    run_coroutine,
    CoroutineMock,
)

try:
    import distro
except ImportError:
    distro = None


class TestVersionServer(unittest.TestCase):
    def setUp(self):
        self.client = make_connected_client()
        self.disco_server = unittest.mock.Mock()
        self.s = version_svc.VersionServer(
            self.client,
            dependencies={
                aioxmpp.disco.DiscoServer: self.disco_server,
            }
        )

    def test_is_service(self):
        self.assertTrue(issubclass(
            version_svc.VersionServer,
            aioxmpp.service.Service,
        ))

    def test_depends_on_DiscoServer(self):
        self.assertIn(
            aioxmpp.disco.DiscoServer,
            version_svc.VersionServer.ORDER_AFTER,
        )

    def test_declares_disco_feature(self):
        self.assertIsInstance(
            version_svc.VersionServer.disco_feature,
            aioxmpp.disco.register_feature,
        )
        self.assertEqual(
            version_svc.VersionServer.disco_feature.feature,
            "jabber:iq:version",
        )

    @unittest.skipIf(distro is not None,
                     "distro package is installed")
    def test_fallback_to_platform_module(self):
        with unittest.mock.patch("platform.system") as system:
            s = version_svc.VersionServer(
                self.client,
                dependencies={
                    aioxmpp.disco.DiscoServer: self.disco_server,
                }
            )

            system.assert_called_once_with()
            system.reset_mock()

            self.assertEqual(s.os, system())

    @unittest.skipIf(distro is None,
                     "distro package not installed")
    def test_try_use_distro_package(self):
        with unittest.mock.patch("distro.name") as name:
            s = version_svc.VersionServer(
                self.client,
                dependencies={
                    aioxmpp.disco.DiscoServer: self.disco_server,
                }
            )

            name.assert_called_once_with()
            name.reset_mock()

            self.assertEqual(s.os, name())

    def test_os_attribute_can_be_set_and_stringifies(self):
        self.s.os = unittest.mock.sentinel.os
        self.assertEqual(self.s.os, str(unittest.mock.sentinel.os))

    def test_os_attribute_can_be_set_to_None(self):
        self.s.os = None
        self.assertIsNone(self.s.os)

    def test_os_attribute_can_be_deleted_to_None(self):
        del self.s.os
        self.assertIsNone(self.s.os)

    def test_name_attribute_init_to_None(self):
        self.assertIsNone(self.s.name)

    def test_name_attribute_can_be_set_and_stringifies(self):
        self.s.name = unittest.mock.sentinel.name
        self.assertEqual(self.s.name, str(unittest.mock.sentinel.name))

    def test_name_attribute_can_be_deleted_to_None(self):
        del self.s.name
        self.assertIsNone(self.s.name)

    def test_version_attribute_init_to_None(self):
        self.assertIsNone(self.s.version)

    def test_version_attribute_can_be_set_and_stringifies(self):
        self.s.version = unittest.mock.sentinel.version
        self.assertEqual(self.s.version, str(unittest.mock.sentinel.version))

    def test_version_attribute_can_be_deleted_to_None(self):
        del self.s.version
        self.assertIsNone(self.s.version)

    def test_handles_query_IQs(self):
        self.assertTrue(
            aioxmpp.service.is_iq_handler(
                aioxmpp.IQType.GET,
                version_xso.Query,
                version_svc.VersionServer.handle_query,
            )
        )

    def test_handler_returns_filled_query_object(self):
        self.s.os = unittest.mock.sentinel.os
        self.s.name = unittest.mock.sentinel.name
        self.s.version = unittest.mock.sentinel.version

        result = run_coroutine(
            self.s.handle_query(unittest.mock.sentinel.request)
        )

        self.assertIsInstance(
            result,
            version_xso.Query,
        )

        self.assertEqual(
            result.os,
            self.s.os,
        )

        self.assertEqual(
            result.name,
            self.s.name,
        )

        self.assertEqual(
            result.version,
            self.s.version,
        )

    def test_handler_returns_unspecified_for_None_version(self):
        self.s.os = unittest.mock.sentinel.os
        self.s.name = unittest.mock.sentinel.name
        self.s.version = None

        result = run_coroutine(
            self.s.handle_query(unittest.mock.sentinel.request)
        )

        self.assertIsInstance(
            result,
            version_xso.Query,
        )

        self.assertEqual(
            result.os,
            self.s.os,
        )

        self.assertEqual(
            result.name,
            self.s.name,
        )

        self.assertEqual(
            result.version,
            "unspecified",
        )

    def test_handler_raises_service_unavailable_if_name_is_unset(self):
        del self.s.name
        self.s.version = "foo"
        self.s.os = "bar"

        with self.assertRaises(aioxmpp.XMPPCancelError) as ctx:
            run_coroutine(self.s.handle_query(unittest.mock.sentinel.request))

        self.assertEqual(
            ctx.exception.condition,
            aioxmpp.ErrorCondition.SERVICE_UNAVAILABLE
        )


class Testquery_version(unittest.TestCase):
    def test_queries_version_over_stream(self):
        test_jid = aioxmpp.JID.fromstr("juliet@capulet.lit")

        stream = unittest.mock.Mock(["send"])
        stream.send = CoroutineMock()
        stream.send.return_value = unittest.mock.sentinel.result

        result = run_coroutine(version_svc.query_version(
            stream,
            test_jid,
        ))

        stream.send.assert_called_once_with(unittest.mock.ANY)

        _, (iq, ), _ = stream.send.mock_calls[0]

        self.assertIsInstance(iq, aioxmpp.IQ)
        self.assertEqual(iq.type_, aioxmpp.IQType.GET)
        self.assertEqual(iq.to, test_jid)

        self.assertIsInstance(iq.payload, version_xso.Query)
        self.assertIsNone(iq.payload.name)
        self.assertIsNone(iq.payload.version)
        self.assertIsNone(iq.payload.os)

        self.assertEqual(
            result,
            unittest.mock.sentinel.result,
        )



########################################################################
# File name: connector.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.connector` --- Ways to establish XML streams
###########################################################

This module provides classes to establish XML streams. Currently, there are two
different ways to establish XML streams: normal TCP connection which is then
upgraded using STARTTLS, and directly using TLS.

.. versionadded:: 0.6

   The whole module was added in version 0.6.

Abstract base class
===================

The connectors share a common abstract base class, :class:`BaseConnector`:

.. autoclass:: BaseConnector

Specific connectors
===================

.. autoclass:: STARTTLSConnector

.. autoclass:: XMPPOverTLSConnector

"""

import abc
import asyncio
import logging

from datetime import timedelta

import aioxmpp.errors as errors
import aioxmpp.nonza as nonza
import aioxmpp.protocol as protocol
import aioxmpp.ssl_transport as ssl_transport

from aioxmpp.utils import namespaces


class BaseConnector(metaclass=abc.ABCMeta):
    """
    This is the base class for connectors. It defines the public interface of
    all connectors.

    .. autoattribute:: tls_supported

    .. automethod:: connect

    Existing connectors:

    .. autosummary::

       STARTTLSConnector
       XMPPOverTLSConnector

    """

    @abc.abstractproperty
    def tls_supported(self):
        """
        Boolean which indicates whether TLS is supported by this connector.
        """

    @abc.abstractproperty
    def dane_supported(self):
        """
        Boolean which indicates whether DANE is supported by this connector.
        """

    @abc.abstractmethod
    @asyncio.coroutine
    def connect(self, loop, metadata, domain, host, port, negotiation_timeout,
                base_logger=None):
        """
        Establish a :class:`.protocol.XMLStream` for `domain` with the given
        `host` at the given TCP `port`.

        `metadata` must be a :class:`.security_layer.SecurityLayer` instance to
        use for the connection. `loop` must be a :class:`asyncio.BaseEventLoop`
        to use.

        `negotiation_timeout` must be the maximum time in seconds to wait for
        the server to reply in each negotiation step. The `negotiation_timeout`
        is used as value for
        :attr:`~aioxmpp.protocol.XMLStream.deadtime_hard_limit` in the returned
        stream.

        Return a triple consisting of the :class:`asyncio.Transport`, the
        :class:`.protocol.XMLStream` and the
        :class:`aioxmpp.nonza.StreamFeatures` of the stream.

        To detect the use of TLS on the stream, check whether
        :meth:`asyncio.Transport.get_extra_info` returns a non-:data:`None`
        value for ``"ssl_object"``.

        `base_logger` is passed to :class:`aioxmpp.protocol.XMLStream`.

        .. versionchanged:: 0.10

            Assignment of
            :attr:`~aioxmpp.protocol.XMLStream.deadtime_hard_limit` was added.
        """


class STARTTLSConnector(BaseConnector):
    """
    Establish an XML stream using STARTTLS.

    .. automethod:: connect
    """

    @property
    def tls_supported(self):
        return True

    @property
    def dane_supported(self):
        return False

    @asyncio.coroutine
    def connect(self, loop, metadata, domain: str, host, port,
                negotiation_timeout, base_logger=None):
        """
        .. seealso::

           :meth:`BaseConnector.connect`
              For general information on the :meth:`connect` method.

        Connect to `host` at TCP port number `port`. The
        :class:`aioxmpp.security_layer.SecurityLayer` object `metadata` is used
        to determine the parameters of the TLS connection.

        First, a normal TCP connection is opened and the stream header is sent.
        The stream features are waited for, and then STARTTLS is negotiated if
        possible.

        :attr:`~.security_layer.SecurityLayer.tls_required` is honoured: if it
        is true and TLS negotiation fails, :class:`~.errors.TLSUnavailable` is
        raised. TLS negotiation is always attempted if
        :attr:`~.security_layer.SecurityLayer.tls_required` is true, even if
        the server does not advertise a STARTTLS stream feature. This might
        help to prevent trivial downgrade attacks, and we don’t have anything
        to lose at this point anymore anyways.

        :attr:`~.security_layer.SecurityLayer.ssl_context_factory` and
        :attr:`~.security_layer.SecurityLayer.certificate_verifier_factory` are
        used to configure the TLS connection.

        .. versionchanged:: 0.10

            The `negotiation_timeout` is set as
            :attr:`~.XMLStream.deadtime_hard_limit` on the returned XML stream.
        """

        features_future = asyncio.Future(loop=loop)

        stream = protocol.XMLStream(
            to=domain,
            features_future=features_future,
            base_logger=base_logger,
        )
        if base_logger is not None:
            logger = base_logger.getChild(type(self).__name__)
        else:
            logger = logging.getLogger(".".join([
                __name__, type(self).__qualname__,
            ]))

        try:
            transport, _ = yield from ssl_transport.create_starttls_connection(
                loop,
                lambda: stream,
                host=host,
                port=port,
                peer_hostname=host,
                server_hostname=domain,
                use_starttls=True,
            )
        except:  # NOQA
            stream.abort()
            raise

        stream.deadtime_hard_limit = timedelta(seconds=negotiation_timeout)

        features = yield from features_future

        try:
            features[nonza.StartTLSFeature]
        except KeyError:
            if not metadata.tls_required:
                return transport, stream, (yield from features_future)
            logger.debug(
                "attempting STARTTLS despite not announced since it is"
                " required")

        try:
            response = yield from protocol.send_and_wait_for(
                stream,
                [
                    nonza.StartTLS(),
                ],
                [
                    nonza.StartTLSFailure,
                    nonza.StartTLSProceed,
                ]
            )
        except errors.StreamError as exc:
            raise errors.TLSUnavailable(
                "STARTTLS not supported by server, but required by client"
            )

        if not isinstance(response, nonza.StartTLSProceed):
            if metadata.tls_required:
                message = (
                    "server failed to STARTTLS"
                )

                protocol.send_stream_error_and_close(
                    stream,
                    condition=errors.StreamErrorCondition.POLICY_VIOLATION,
                    text=message,
                )

                raise errors.TLSUnavailable(message)
            return transport, stream, (yield from features_future)

        verifier = metadata.certificate_verifier_factory()
        yield from verifier.pre_handshake(
            domain,
            host,
            port,
            metadata,
        )

        ssl_context = metadata.ssl_context_factory()
        verifier.setup_context(ssl_context, transport)

        yield from stream.starttls(
            ssl_context=ssl_context,
            post_handshake_callback=verifier.post_handshake,
        )

        features_future = yield from protocol.reset_stream_and_get_features(
            stream,
            timeout=negotiation_timeout,
        )

        return transport, stream, features_future


class XMPPOverTLSConnector(BaseConnector):
    """
    Establish an XML stream using XMPP-over-TLS, as per :xep:`368`.

    .. automethod:: connect
    """

    @property
    def dane_supported(self):
        return False

    @property
    def tls_supported(self):
        return True

    def _context_factory_factory(self, logger, metadata, verifier):
        def context_factory(transport):
            ssl_context = metadata.ssl_context_factory()

            if hasattr(ssl_context, "set_alpn_protos"):
                try:
                    ssl_context.set_alpn_protos([b'xmpp-client'])
                except NotImplementedError:
                    logger.warning(
                        "the underlying OpenSSL library does not support ALPN"
                    )
            else:
                logger.warning(
                    "OpenSSL.SSL.Context lacks set_alpn_protos - "
                    "please update pyOpenSSL to a recent version"
                )

            verifier.setup_context(ssl_context, transport)
            return ssl_context
        return context_factory

    @asyncio.coroutine
    def connect(self, loop, metadata, domain, host, port,
                negotiation_timeout, base_logger=None):
        """
        .. seealso::

           :meth:`BaseConnector.connect`
              For general information on the :meth:`connect` method.

        Connect to `host` at TCP port number `port`. The
        :class:`aioxmpp.security_layer.SecurityLayer` object `metadata` is used
        to determine the parameters of the TLS connection.

        The connector connects to the server by directly establishing TLS; no
        XML stream is started before TLS negotiation, in accordance to
        :xep:`368` and how legacy SSL was handled in the past.

        :attr:`~.security_layer.SecurityLayer.ssl_context_factory` and
        :attr:`~.security_layer.SecurityLayer.certificate_verifier_factory` are
        used to configure the TLS connection.

        .. versionchanged:: 0.10

            The `negotiation_timeout` is set as
            :attr:`~.XMLStream.deadtime_hard_limit` on the returned XML stream.
        """

        features_future = asyncio.Future(loop=loop)

        stream = protocol.XMLStream(
            to=domain,
            features_future=features_future,
            base_logger=base_logger,
        )

        if base_logger is not None:
            logger = base_logger.getChild(type(self).__name__)
        else:
            logger = logging.getLogger(".".join([
                __name__, type(self).__qualname__,
            ]))

        verifier = metadata.certificate_verifier_factory()
        yield from verifier.pre_handshake(
            domain,
            host,
            port,
            metadata,
        )

        context_factory = self._context_factory_factory(logger, metadata,
                                                        verifier)

        try:
            transport, _ = yield from ssl_transport.create_starttls_connection(
                loop,
                lambda: stream,
                host=host,
                port=port,
                peer_hostname=host,
                server_hostname=domain,
                post_handshake_callback=verifier.post_handshake,
                ssl_context_factory=context_factory,
                use_starttls=False,
            )
        except:  # NOQA
            stream.abort()
            raise

        stream.deadtime_hard_limit = timedelta(seconds=negotiation_timeout)

        return transport, stream, (yield from features_future)



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp
import aioxmpp.version.xso as version_xso
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces


class TestNamespaces(unittest.TestCase):
    def test_namespace(self):
        self.assertEqual(
            namespaces.xep0092_version,
            "jabber:iq:version",
        )


class TestQuery(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            version_xso.Query,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            version_xso.Query.TAG,
            (namespaces.xep0092_version, "query"),
        )

    def test_version(self):
        self.assertIsInstance(
            version_xso.Query.version.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            version_xso.Query.version.tag,
            (namespaces.xep0092_version, "version")
        )
        self.assertIsNone(
            version_xso.Query.version.default,
        )

    def test_name(self):
        self.assertIsInstance(
            version_xso.Query.name.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            version_xso.Query.name.tag,
            (namespaces.xep0092_version, "name")
        )
        self.assertIsNone(
            version_xso.Query.name.default,
        )

    def test_os(self):
        self.assertIsInstance(
            version_xso.Query.os.xq_descriptor,
            xso.ChildText,
        )
        self.assertEqual(
            version_xso.Query.os.tag,
            (namespaces.xep0092_version, "os")
        )
        self.assertIsNone(
            version_xso.Query.os.default,
        )

    def test_is_iq_payload(self):
        self.assertIn(
            version_xso.Query.TAG,
            aioxmpp.IQ.CHILD_MAP,
        )



########################################################################
# File name: test_markers.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp.misc as misc_xso
import aioxmpp.misc.markers as markers_xso
import aioxmpp.xso

from aioxmpp.utils import namespaces


class TestNamespaces(unittest.TestCase):
    def test_namespace(self):
        self.assertEqual(
            namespaces.xep0333_markers,
            "urn:xmpp:chat-markers:0",
        )


class TestMarker(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            markers_xso.Marker,
            aioxmpp.xso.XSO
        ))

    def test_id_(self):
        self.assertIsInstance(
            markers_xso.Marker.id_,
            aioxmpp.xso.Attr,
        )
        self.assertEqual(
            markers_xso.Marker.id_.tag,
            (None, "id")
        )


class TestReceived(unittest.TestCase):
    def test_is_marker(self):
        self.assertTrue(issubclass(
            misc_xso.ReceivedMarker,
            markers_xso.Marker,
        ))

    def test_tag(self):
        self.assertEqual(
            misc_xso.ReceivedMarker.TAG,
            (namespaces.xep0333_markers, "received"),
        )


class TestDisplayed(unittest.TestCase):
    def test_is_marker(self):
        self.assertTrue(issubclass(
            misc_xso.DisplayedMarker,
            markers_xso.Marker,
        ))

    def test_tag(self):
        self.assertEqual(
            misc_xso.DisplayedMarker.TAG,
            (namespaces.xep0333_markers, "displayed"),
        )


class TestAcknowledged(unittest.TestCase):
    def test_is_marker(self):
        self.assertTrue(issubclass(
            misc_xso.AcknowledgedMarker,
            markers_xso.Marker,
        ))

    def test_tag(self):
        self.assertEqual(
            misc_xso.AcknowledgedMarker.TAG,
            (namespaces.xep0333_markers, "acknowledged"),
        )


class TestMessage(unittest.TestCase):
    def test_xep0333_marker(self):
        self.assertIsInstance(
            aioxmpp.Message.xep0333_marker,
            aioxmpp.xso.Child,
        )
        self.assertLessEqual(
            {
                misc_xso.ReceivedMarker,
                misc_xso.DisplayedMarker,
                misc_xso.AcknowledgedMarker,
            },
            set(aioxmpp.Message.xep0333_marker._classes)
        )



#!/usr/bin/env python3
########################################################################
# File name: setup.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import os.path
import runpy
import sys

import setuptools
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

version_mod = runpy.run_path("aioxmpp/_version.py")

install_requires = [
    'aiosasl>=0.3',  # need 0.2+ for LGPLv3
    'aioopenssl>=0.1',
    'babel~=2.3',
    'dnspython~=1.0',
    'lxml~=4.0',
    'multidict<5,>=2.0',
    'sortedcollections>=0.5',
    'pyOpenSSL',
    'pyasn1',
    'pyasn1_modules',
    'tzlocal~=1.2',
]

if tuple(map(int, setuptools.__version__.split(".")[:3])) < (6, 0, 0):
    for i, item in enumerate(install_requires):
        install_requires[i] = item.replace("~=", ">=")

if sys.version_info[:3] < (3, 5, 0):
    install_requires.append("typing")

setup(
    name="aioxmpp",
    version=version_mod["__version__"].replace("-", ""),
    description="Pure-python XMPP library for asyncio",
    long_description=long_description,
    url="https://github.com/horazont/aioxmpp",
    author="Jonas Wielicki",
    author_email="jonas@wielicki.name",
    license="LGPLv3+",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Operating System :: POSIX",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Topic :: Communications :: Chat",
        "Topic :: Internet :: XMPP",
    ],
    keywords="asyncio xmpp library",
    install_requires=install_requires,
    packages=find_packages(exclude=["tests*", "benchmarks*"])
)



########################################################################
# File name: test_callbacks.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import contextlib
import functools
import unittest
import unittest.mock

from aioxmpp.callbacks import (
    TagDispatcher,
    TagListener,
    AsyncTagListener,
    OneshotTagListener,
    OneshotAsyncTagListener,
    FutureListener,
    Signal,
    AdHocSignal,
    SyncAdHocSignal,
    SyncSignal,
    Filter,
    first_signal,
)

from aioxmpp.testutils import run_coroutine, CoroutineMock


class TestTagListener(unittest.TestCase):
    def test_data(self):
        ondata = unittest.mock.Mock()

        obj = object()

        listener = TagListener(ondata=ondata)
        listener.data(obj)
        ondata.assert_called_once_with(obj)

    def test_uninitialized_error(self):
        ondata = unittest.mock.Mock()

        listener = TagListener(ondata=ondata)
        listener.error(ValueError())

    def test_error(self):
        ondata = unittest.mock.Mock()
        onerror = unittest.mock.Mock()

        exc = ValueError()

        listener = TagListener(ondata, onerror)
        listener.error(exc)

        ondata.assert_not_called()
        onerror.assert_called_once_with(exc)

    def test_is_valid(self):
        self.assertTrue(TagListener(ondata=unittest.mock.Mock()))


class TestTagDispatcher(unittest.TestCase):
    def test_add_callback(self):
        mock = unittest.mock.Mock()

        nh = TagDispatcher()
        nh.add_callback("tag", mock)
        with self.assertRaisesRegex(ValueError,
                                     "only one listener is allowed"):
            nh.add_callback("tag", mock)

    def test_add_listener(self):
        mock = unittest.mock.Mock()

        l = TagListener(mock)

        nh = TagDispatcher()
        nh.add_listener("tag", l)
        with self.assertRaisesRegex(ValueError,
                                     "only one listener is allowed"):
            nh.add_listener("tag", l)

    def test_add_listener_skips_invalid(self):
        mock = unittest.mock.Mock()

        l1 = unittest.mock.Mock()
        l1.is_valid.return_value = True

        l2 = TagListener(mock)

        nh = TagDispatcher()
        nh.add_listener("tag", l1)
        l1.is_valid.return_value = False
        nh.add_listener("tag", l2)

        obj = object()
        nh.unicast("tag", obj)
        self.assertSequenceEqual(
            [
                unittest.mock.call.is_valid(),
            ],
            l1.mock_calls
        )

        self.assertSequenceEqual(
            [
                unittest.mock.call(obj),
            ],
            mock.mock_calls
        )

    @unittest.mock.patch("aioxmpp.callbacks.AsyncTagListener")
    def test_add_callback_async(self, AsyncTagListener):
        AsyncTagListener().is_valid.return_value = True
        AsyncTagListener.mock_calls.clear()

        data = unittest.mock.Mock()
        loop = unittest.mock.Mock()
        obj = object()

        nh = TagDispatcher()
        nh.add_callback_async("tag", data, loop=loop)

        self.assertSequenceEqual(
            [
                unittest.mock.call(data, loop=loop)
            ],
            AsyncTagListener.mock_calls
        )
        del AsyncTagListener.mock_calls[:]

        nh.unicast("tag", obj)

        self.assertSequenceEqual(
            [
                unittest.mock.call().is_valid(),
                unittest.mock.call().data(obj),
                unittest.mock.call().data().__bool__(),
            ],
            AsyncTagListener.mock_calls
        )

    def test_add_future(self):
        mock = unittest.mock.Mock()
        mock.done.return_value = False
        obj = object()

        nh = TagDispatcher()
        nh.add_future("tag", mock)
        nh.unicast("tag", obj)
        with self.assertRaises(KeyError):
            # futures must be oneshots
            nh.unicast("tag", obj)

        nh.add_future("tag", mock)
        nh.broadcast_error(obj)
        with self.assertRaises(KeyError):
            # futures must be oneshots
            nh.unicast("tag", obj)

        self.assertSequenceEqual(
            [
                unittest.mock.call.done(),
                unittest.mock.call.set_result(obj),
                unittest.mock.call.done(),
                unittest.mock.call.set_exception(obj),
            ],
            mock.mock_calls
        )

    def test_unicast(self):
        mock = unittest.mock.Mock()
        mock.return_value = False
        obj = object()

        nh = TagDispatcher()
        nh.add_callback("tag", mock)
        nh.unicast("tag", obj)
        nh.unicast("tag", obj)

        self.assertSequenceEqual(
            [
                unittest.mock.call(obj),
                unittest.mock.call(obj),
            ],
            mock.mock_calls
        )

    def test_unicast_fails_for_nonexistent(self):
        obj = object()
        nh = TagDispatcher()
        with self.assertRaises(KeyError):
            nh.unicast("tag", obj)

    def test_unicast_fails_for_invalid(self):
        obj = object()
        l = unittest.mock.Mock()
        l.is_valid.return_value = False
        nh = TagDispatcher()
        nh.add_listener("tag", l)
        with self.assertRaises(KeyError):
            nh.unicast("tag", obj)

    def test_unicast_to_oneshot(self):
        mock = unittest.mock.Mock()
        obj = object()

        l = OneshotTagListener(mock)

        nh = TagDispatcher()
        nh.add_listener("tag", l)

        nh.unicast("tag", obj)
        with self.assertRaises(KeyError):
            nh.unicast("tag", obj)

        self.assertSequenceEqual(
            [
                unittest.mock.call(obj)
            ],
            mock.mock_calls
        )

    def test_unicast_to_cancelled_oneshot(self):
        mock = unittest.mock.Mock()
        obj = object()

        l = OneshotTagListener(mock)

        nh = TagDispatcher()
        nh.add_listener("tag", l)

        l.cancel()

        with self.assertRaises(KeyError):
            nh.unicast("tag", obj)

        self.assertSequenceEqual(
            [],
            mock.mock_calls
        )

    def test_unicast_removes_for_true_result(self):
        mock = unittest.mock.Mock()
        mock.return_value = True
        obj = object()

        nh = TagDispatcher()
        nh.add_callback("tag", mock)
        nh.unicast("tag", obj)
        with self.assertRaises(KeyError):
            nh.unicast("tag", obj)

        mock.assert_called_once_with(obj)

    def test_broadcast_error_to_oneshot(self):
        data = unittest.mock.Mock()
        error = unittest.mock.Mock()
        obj = object()

        l = OneshotTagListener(data, error)

        nh = TagDispatcher()
        nh.add_listener("tag", l)

        nh.broadcast_error(obj)
        with self.assertRaises(KeyError):
            nh.unicast("tag", obj)

        self.assertSequenceEqual(
            [
                unittest.mock.call(obj)
            ],
            error.mock_calls
        )
        self.assertFalse(data.mock_calls)

    def test_broadcast_error_to_cancelled_oneshot(self):
        data = unittest.mock.Mock()
        error = unittest.mock.Mock()
        obj = object()

        l = OneshotTagListener(data, error)

        nh = TagDispatcher()
        nh.add_listener("tag", l)

        l.cancel()

        nh.broadcast_error(obj)
        with self.assertRaises(KeyError):
            nh.unicast("tag", obj)

        self.assertSequenceEqual(
            [],
            error.mock_calls
        )
        self.assertFalse(data.mock_calls)

    def test_broadcast_error_skip_invalid(self):
        obj = object()
        l = unittest.mock.Mock()
        l.is_valid.return_value = False
        nh = TagDispatcher()
        nh.add_listener("tag", l)
        nh.broadcast_error(obj)
        self.assertSequenceEqual(
            [
                unittest.mock.call.is_valid()
            ],
            l.mock_calls
        )

    def test_remove_listener(self):
        mock = unittest.mock.Mock()
        nh = TagDispatcher()
        nh.add_callback("tag", mock)
        nh.remove_listener("tag")
        with self.assertRaises(KeyError):
            nh.unicast("tag", object())
        mock.assert_not_called()

    def test_broadcast_error(self):
        data = unittest.mock.Mock()
        error1 = unittest.mock.Mock()
        error1.return_value = False
        error2 = unittest.mock.Mock()
        error2.return_value = False

        l1 = TagListener(data, error1)
        l2 = TagListener(data, error2)

        obj = object()

        nh = TagDispatcher()
        nh.add_listener("tag1", l1)
        nh.add_listener("tag2", l2)

        nh.broadcast_error(obj)
        nh.broadcast_error(obj)

        data.assert_not_called()
        self.assertSequenceEqual(
            [
                unittest.mock.call(obj),
                unittest.mock.call(obj),
            ],
            error1.mock_calls
        )
        self.assertSequenceEqual(
            [
                unittest.mock.call(obj),
                unittest.mock.call(obj),
            ],
            error2.mock_calls
        )

    def test_broadcast_error_removes_on_true_result(self):
        data = unittest.mock.Mock()
        error1 = unittest.mock.Mock()
        error1.return_value = True

        l1 = TagListener(data, error1)

        obj = object()

        nh = TagDispatcher()
        nh.add_listener("tag1", l1)

        nh.broadcast_error(obj)
        nh.broadcast_error(obj)

        data.assert_not_called()
        self.assertSequenceEqual(
            [
                unittest.mock.call(obj),
            ],
            error1.mock_calls
        )

    def test_unicast_error(self):
        data = unittest.mock.Mock()
        error1 = unittest.mock.Mock()
        error1.return_value = False
        error2 = unittest.mock.Mock()
        error2.return_value = False

        l1 = TagListener(data, error1)
        l2 = TagListener(data, error2)

        obj = object()

        nh = TagDispatcher()
        nh.add_listener("tag1", l1)
        nh.add_listener("tag2", l2)

        nh.unicast_error("tag1", obj)

        data.assert_not_called()
        self.assertSequenceEqual(
            [
                unittest.mock.call(obj),
            ],
            error1.mock_calls
        )
        self.assertSequenceEqual(
            [
            ],
            error2.mock_calls
        )

    def test_unicast_error_skip_and_remove_invalid_and_raise_KeyError(self):
        obj = object()
        l = unittest.mock.Mock()
        l.is_valid.return_value = False
        nh = TagDispatcher()
        nh.add_listener("tag", l)
        with self.assertRaises(KeyError):
            nh.unicast_error("tag", obj)

        self.assertSequenceEqual(
            [
                unittest.mock.call.is_valid()
            ],
            l.mock_calls
        )

    def test_unicast_error_removes_on_true_result(self):
        data = unittest.mock.Mock()
        error1 = unittest.mock.Mock()
        error1.return_value = True

        l1 = TagListener(data, error1)

        obj = object()

        nh = TagDispatcher()
        nh.add_listener("tag1", l1)

        nh.unicast_error("tag1", obj)
        with self.assertRaises(KeyError):
            nh.unicast_error("tag1", obj)

        data.assert_not_called()
        self.assertSequenceEqual(
            [
                unittest.mock.call(obj),
            ],
            error1.mock_calls
        )

    def test_close(self):
        data = unittest.mock.Mock()
        error1 = unittest.mock.Mock()
        error2 = unittest.mock.Mock()

        l1 = TagListener(data, error1)
        l2 = TagListener(data, error2)

        obj = object()

        nh = TagDispatcher()
        nh.add_listener("tag1", l1)
        nh.add_listener("tag2", l2)

        nh.close_all(obj)

        data.assert_not_called()
        error1.assert_called_once_with(obj)
        error2.assert_called_once_with(obj)

        with self.assertRaises(KeyError):
            nh.remove_listener("tag1")
        with self.assertRaises(KeyError):
            nh.remove_listener("tag2")
        with self.assertRaises(KeyError):
            nh.unicast("tag1", None)
        with self.assertRaises(KeyError):
            nh.unicast("tag2", None)


class TestAsyncTagListener(unittest.TestCase):
    def test_everything(self):
        data = unittest.mock.MagicMock()
        error = unittest.mock.MagicMock()
        loop = unittest.mock.MagicMock()
        obj = object()
        tl = AsyncTagListener(data, error, loop=loop)
        self.assertFalse(tl.data(obj))
        self.assertFalse(tl.error(obj))

        self.assertFalse(data.mock_calls)
        self.assertFalse(error.mock_calls)
        self.assertSequenceEqual(
            [
                unittest.mock.call.__bool__(),
                unittest.mock.call.call_soon(data, obj),
                unittest.mock.call.call_soon(error, obj),
            ],
            loop.mock_calls
        )


class TestOneshotAsyncTagListener(unittest.TestCase):
    def test_everything(self):
        data = unittest.mock.MagicMock()
        error = unittest.mock.MagicMock()
        loop = unittest.mock.MagicMock()
        obj = object()
        tl = OneshotAsyncTagListener(data, error, loop=loop)
        self.assertTrue(tl.data(obj))
        self.assertTrue(tl.error(obj))

        self.assertFalse(data.mock_calls)
        self.assertFalse(error.mock_calls)
        self.assertSequenceEqual(
            [
                unittest.mock.call.__bool__(),
                unittest.mock.call.call_soon(data, obj),
                unittest.mock.call.call_soon(error, obj),
            ],
            loop.mock_calls
        )


class TestFutureListener(unittest.TestCase):
    def test_normal_operation(self):
        loop = asyncio.get_event_loop()
        fut = asyncio.Future(loop=loop)
        obj = object()
        tl = FutureListener(fut)

        self.assertTrue(tl.is_valid())

        self.assertTrue(tl.data(obj))
        self.assertEqual(fut.result(), obj)

        self.assertFalse(tl.is_valid())

    def test_error_dispatch(self):
        loop = asyncio.get_event_loop()
        fut = asyncio.Future(loop=loop)
        obj = Exception()
        tl = FutureListener(fut)

        self.assertTrue(tl.is_valid())

        self.assertTrue(tl.error(obj))
        self.assertEqual(fut.exception(), obj)

        self.assertFalse(tl.is_valid())

    def test_signals_non_existance_with_cancelled_future(self):
        loop = asyncio.get_event_loop()
        fut = asyncio.Future(loop=loop)
        tl = FutureListener(fut)

        self.assertTrue(tl.is_valid())

        fut.cancel()

        self.assertFalse(tl.is_valid())

    def test_swallow_invalid_state_error(self):
        loop = asyncio.get_event_loop()
        fut = asyncio.Future(loop=loop)
        obj = object()
        tl = FutureListener(fut)

        fut.cancel()

        self.assertTrue(tl.data(obj))
        self.assertTrue(tl.error(obj))


class TestAdHocSignal(unittest.TestCase):
    def test_STRONG_rejects_non_callable(self):
        signal = AdHocSignal()
        with self.assertRaisesRegex(TypeError, "must be callable"):
            signal.STRONG(object())

    def test_WEAK_rejects_non_callable(self):
        signal = AdHocSignal()
        with self.assertRaisesRegex(TypeError, "must be callable"):
            signal.WEAK(object())

    def test_ASYNC_WITH_LOOP_rejects_non_callable(self):
        signal = AdHocSignal()
        with self.assertRaisesRegex(TypeError, "must be callable"):
            signal.ASYNC_WITH_LOOP(asyncio.get_event_loop())(object())

    def test_connect_and_fire(self):
        signal = AdHocSignal()

        fun = unittest.mock.MagicMock()
        fun.return_value = None

        signal.connect(fun)

        signal.fire()
        signal.fire()

        self.assertSequenceEqual(
            [
                unittest.mock.call(),
                unittest.mock.call(),
            ],
            fun.mock_calls
        )

    def test_connect_and_call(self):
        signal = AdHocSignal()

        fun = unittest.mock.MagicMock()
        fun.return_value = None

        signal.connect(fun)

        signal()

        fun.assert_called_once_with()

    def test_connect_weak_uses_weakref(self):
        signal = AdHocSignal()

        with unittest.mock.patch("weakref.ref") as ref:
            fun = unittest.mock.MagicMock()
            signal.connect(fun, AdHocSignal.WEAK)
            ref.assert_called_once_with(fun)

    def test_connect_weak_uses_WeakMethod_for_methods(self):
        signal = AdHocSignal()

        class Foo:
            def meth(self):
                return None

        f = Foo()

        with unittest.mock.patch("weakref.WeakMethod") as ref:
            signal.connect(f.meth, AdHocSignal.WEAK)

        ref.assert_called_once_with(f.meth)

    def test_connect_does_not_use_weakref(self):
        signal = AdHocSignal()

        with unittest.mock.patch("weakref.ref") as ref:
            fun = unittest.mock.MagicMock()
            signal.connect(fun)
            self.assertFalse(ref.mock_calls)

    @unittest.mock.patch("weakref.ref")
    def test_fire_removes_stale_references(self, ref):
        signal = AdHocSignal()

        fun = unittest.mock.MagicMock()
        fun.return_value = None
        ref().return_value = None

        signal.connect(fun, AdHocSignal.WEAK)

        signal.fire()

        self.assertFalse(signal._connections)

    def test_connect_async(self):
        signal = AdHocSignal()

        mock = unittest.mock.MagicMock()
        fun = functools.partial(mock)

        signal.connect(fun, AdHocSignal.ASYNC_WITH_LOOP(None))
        signal.fire()

        mock.assert_not_called()

        run_coroutine(asyncio.sleep(0))

        mock.assert_called_once_with()

    def test_connect_spawn(self):
        signal = AdHocSignal()

        mock = CoroutineMock()

        @asyncio.coroutine
        def coro(*args, **kwargs):
            yield from mock(*args, **kwargs)

        signal.connect(coro, AdHocSignal.SPAWN_WITH_LOOP(None))
        signal.fire("a", 1, b="c")

        self.assertSequenceEqual(mock.mock_calls, [])

        run_coroutine(asyncio.sleep(0))

        self.assertSequenceEqual(
            mock.mock_calls,
            [
                unittest.mock.call("a", 1, b="c")
            ]
        )

    def test_connect_spawn_emits_always(self):
        signal = AdHocSignal()

        mock = CoroutineMock()

        @asyncio.coroutine
        def coro(*args, **kwargs):
            yield from mock(*args, **kwargs)

        signal.connect(coro, AdHocSignal.SPAWN_WITH_LOOP(None))
        signal.fire("a", 1, b="c")
        signal.fire("x")

        self.assertSequenceEqual(mock.mock_calls, [])

        run_coroutine(asyncio.sleep(0))

        run_coroutine(asyncio.sleep(0))

        self.assertSequenceEqual(
            mock.mock_calls,
            [
                unittest.mock.call("a", 1, b="c"),
                unittest.mock.call("x"),
            ]
        )

    def test_SPAWN_rejects_non_coroutine(self):
        def fun():
            pass

        signal = AdHocSignal()

        with self.assertRaisesRegex(TypeError, "must be coroutine"):
            signal.SPAWN_WITH_LOOP(None)(fun)

    def test_fire_with_arguments(self):
        signal = AdHocSignal()

        fun = unittest.mock.MagicMock()

        signal.connect(fun)

        signal("a", 1, foo=None)

        fun.assert_called_once_with("a", 1, foo=None)

    def test_remove_callback_on_true_result(self):
        signal = AdHocSignal()

        fun = unittest.mock.MagicMock()
        fun.return_value = True

        signal.connect(fun)

        signal()

        self.assertSequenceEqual(
            [
                unittest.mock.call(),
            ],
            fun.mock_calls
        )

        signal()

        self.assertSequenceEqual(
            [
                unittest.mock.call(),
            ],
            fun.mock_calls
        )

    def test_remove_by_token(self):
        signal = AdHocSignal()

        fun = unittest.mock.MagicMock()
        fun.return_value = None

        token = signal.connect(fun)

        signal()

        self.assertSequenceEqual(
            [
                unittest.mock.call(),
            ],
            fun.mock_calls
        )

        signal()

        self.assertSequenceEqual(
            [
                unittest.mock.call(),
                unittest.mock.call(),
            ],
            fun.mock_calls
        )

        signal.disconnect(token)

        signal()

        self.assertSequenceEqual(
            [
                unittest.mock.call(),
                unittest.mock.call(),
            ],
            fun.mock_calls
        )

    def test_disconnect_is_idempotent(self):
        signal = AdHocSignal()

        fun = unittest.mock.MagicMock()
        fun.return_value = None

        token = signal.connect(fun)

        signal.disconnect(token)
        signal.disconnect(token)

    def test_context_connect(self):
        signal = AdHocSignal()

        fun = unittest.mock.MagicMock()
        fun.return_value = None

        with signal.context_connect(fun):
            signal("foo")
        signal("bar")
        with signal.context_connect(fun) as token:
            signal("baz")
            signal.disconnect(token)
            signal("fnord")

        self.assertSequenceEqual(
            [
                unittest.mock.call("foo"),
                unittest.mock.call("baz"),
            ],
            fun.mock_calls
        )

    def test_context_connect_forwards_exceptions_and_disconnects(self):
        signal = AdHocSignal()

        fun = unittest.mock.MagicMock()
        fun.return_value = None

        exc = ValueError()
        with self.assertRaises(ValueError) as ctx:
            with signal.context_connect(fun):
                signal("foo")
                raise exc
        signal("bar")

        self.assertIs(exc, ctx.exception)

        self.assertSequenceEqual(
            [
                unittest.mock.call("foo"),
            ],
            fun.mock_calls
        )

    def test_connect_auto_future_uses_set_result_with_None(self):
        signal = AdHocSignal()

        fut = unittest.mock.Mock()
        fut.done.return_value = False

        signal.connect(fut, AdHocSignal.AUTO_FUTURE)

        signal()
        signal()

        self.assertSequenceEqual(
            [
                unittest.mock.call.done(),
                unittest.mock.call.set_result(None)
            ],
            fut.mock_calls
        )

    def test_connect_auto_future_uses_set_result_with_argument(self):
        signal = AdHocSignal()

        obj = object()

        fut = unittest.mock.Mock()
        fut.done.return_value = False

        signal.connect(fut, AdHocSignal.AUTO_FUTURE)

        signal(obj)
        signal(obj)

        self.assertSequenceEqual(
            [
                unittest.mock.call.done(),
                unittest.mock.call.set_result(obj)
            ],
            fut.mock_calls
        )

    def test_connect_auto_future_fails_if_more_than_one_argument(self):
        fut = unittest.mock.Mock()
        fut.done.return_value = False

        obj = object()

        wrapped = AdHocSignal.AUTO_FUTURE(fut)

        with self.assertRaises(TypeError):
            wrapped(obj, "foo")

        with self.assertRaises(TypeError):
            wrapped(obj, fnord="foo")

    def test_connect_auto_future_converts_exceptions(self):
        signal = AdHocSignal()

        obj = ValueError()

        fut = unittest.mock.Mock()
        fut.done.return_value = False

        signal.connect(fut, AdHocSignal.AUTO_FUTURE)

        signal(obj)
        signal(obj)

        self.assertSequenceEqual(
            [
                unittest.mock.call.done(),
                unittest.mock.call.set_exception(obj)
            ],
            fut.mock_calls
        )

    def test_connect_skips_if_future_is_done(self):
        signal = AdHocSignal()

        obj = ValueError()

        fut = unittest.mock.Mock()
        fut.done.return_value = True

        signal.connect(fut, AdHocSignal.AUTO_FUTURE)

        signal("foo")
        signal(obj)

        self.assertSequenceEqual(
            [
                unittest.mock.call.done(),
            ],
            fut.mock_calls
        )

    def test_logger_on_connect(self):
        logger = unittest.mock.Mock()
        signal = AdHocSignal()
        signal.logger = logger

        a = unittest.mock.Mock()
        signal.connect(a)

        logger.debug.assert_called_with(
            "connecting %r with mode %r",
            a, signal.STRONG)

    def test_logger_on_emit_with_exception(self):
        logger = unittest.mock.Mock()
        signal = AdHocSignal()
        signal.logger = logger

        a = unittest.mock.Mock()
        signal.connect(a)

        a.side_effect = Exception()

        signal()

        logger.exception.assert_called_with(
            "listener attached to signal raised"
        )

    def test_full_isolation(self):
        signal = AdHocSignal()

        base = unittest.mock.Mock()

        base.a.return_value = None
        base.a.side_effect = OSError()

        base.b.return_value = None
        base.b.side_effect = Exception()

        base.c.return_value = None

        base.d.return_value = None
        base.d.side_effect = ValueError()

        signal.connect(base.a)
        signal.connect(base.b)
        signal.connect(base.c)
        signal.connect(base.d)

        signal()

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.a(),
                unittest.mock.call.b(),
                unittest.mock.call.c(),
                unittest.mock.call.d(),
            ]
        )

        signal()

        self.assertSequenceEqual(
            base.mock_calls,
            [
                unittest.mock.call.a(),
                unittest.mock.call.b(),
                unittest.mock.call.c(),
                unittest.mock.call.d(),
                unittest.mock.call.c(),
            ]
        )

    def test_future(self):
        signal = AdHocSignal()

        with contextlib.ExitStack() as stack:
            Future = stack.enter_context(
                unittest.mock.patch("asyncio.Future")
            )

            connect = stack.enter_context(
                unittest.mock.patch.object(
                    signal,
                    "connect"
                )
            )

            fut = signal.future()

        self.assertSequenceEqual(
            Future.mock_calls,
            [
                unittest.mock.call.Future()
            ]
        )

        self.assertSequenceEqual(
            connect.mock_calls,
            [
                unittest.mock.call(
                    Future(),
                    signal.AUTO_FUTURE,
                ),
            ]
        )

        self.assertEqual(fut, Future())


class TestSyncAdHocSignal(unittest.TestCase):
    def test_connect_and_fire(self):
        coro = CoroutineMock()
        coro.return_value = True

        signal = SyncAdHocSignal()
        signal.connect(coro)

        run_coroutine(signal.fire(1, 2, foo="bar"))

        self.assertSequenceEqual(
            [
                unittest.mock.call(1, 2, foo="bar"),
            ],
            coro.mock_calls
        )

    def test_fire_removes_on_false_result(self):
        coro = CoroutineMock()
        coro.return_value = False

        signal = SyncAdHocSignal()
        signal.connect(coro)

        run_coroutine(signal.fire(1, 2, foo="bar"))

        self.assertSequenceEqual(
            [
                unittest.mock.call(1, 2, foo="bar"),
            ],
            coro.mock_calls
        )
        coro.reset_mock()

        run_coroutine(signal.fire(1, 2, foo="bar"))

        self.assertSequenceEqual(
            [
            ],
            coro.mock_calls
        )

    def test_ordered_calls(self):
        calls = []

        def make_coro(i):
            @asyncio.coroutine
            def coro():
                nonlocal calls
                calls.append(i)
            return coro

        coros = [make_coro(i) for i in range(3)]

        signal = SyncAdHocSignal()
        for coro in reversed(coros):
            signal.connect(coro)

        run_coroutine(signal.fire())

        self.assertSequenceEqual(
            [2, 1, 0],
            calls
        )

    def test_context_connect(self):
        signal = SyncAdHocSignal()

        coro = CoroutineMock()
        coro.return_value = True

        with signal.context_connect(coro):
            run_coroutine(signal("foo"))
        run_coroutine(signal("bar"))
        with signal.context_connect(coro) as token:
            run_coroutine(signal("baz"))
            signal.disconnect(token)
            run_coroutine(signal("fnord"))

        self.assertSequenceEqual(
            [
                unittest.mock.call("foo"),
                unittest.mock.call("baz"),
            ],
            coro.mock_calls
        )


class TestSignal(unittest.TestCase):
    def test_get(self):
        class Foo:
            s = Signal()

        instance1 = Foo()
        instance2 = Foo()

        self.assertIsNot(instance1.s, instance2.s)
        self.assertIs(instance1.s, instance1.s)
        self.assertIs(instance2.s, instance2.s)

        self.assertIsInstance(instance1.s, AdHocSignal)
        self.assertIsInstance(instance2.s, AdHocSignal)

    def test_reject_set(self):
        class Foo:
            s = Signal()

        instance = Foo()

        with self.assertRaises(AttributeError):
            instance.s = "foo"

    def test_reject_delete(self):
        class Foo:
            s = Signal()

        instance = Foo()

        with self.assertRaises(AttributeError):
            del instance.s

    def test_default_docstring(self):
        class Foo:
            s = Signal()

        self.assertIsNone(Foo.s.__doc__)

    def test_set_docstring(self):
        class Foo:
            s = Signal(doc=unittest.mock.sentinel.doc)

        self.assertIs(Foo.s.__doc__, unittest.mock.sentinel.doc)


class TestSyncSignal(unittest.TestCase):
    def test_get(self):
        class Foo:
            s = SyncSignal()

        instance1 = Foo()
        instance2 = Foo()

        self.assertIsNot(instance1.s, instance2.s)
        self.assertIs(instance1.s, instance1.s)
        self.assertIs(instance2.s, instance2.s)

        self.assertIsInstance(instance1.s, SyncAdHocSignal)
        self.assertIsInstance(instance2.s, SyncAdHocSignal)

    def test_reject_set(self):
        class Foo:
            s = SyncSignal()

        instance = Foo()

        with self.assertRaises(AttributeError):
            instance.s = "foo"

    def test_reject_delete(self):
        class Foo:
            s = SyncSignal()

        instance = Foo()

        with self.assertRaises(AttributeError):
            del instance.s

    def test_default_docstring(self):
        class Foo:
            s = SyncSignal()

        self.assertIsNone(Foo.s.__doc__)

    def test_set_docstring(self):
        class Foo:
            s = SyncSignal(doc=unittest.mock.sentinel.doc)

        self.assertIs(Foo.s.__doc__, unittest.mock.sentinel.doc)


class TestFilter_Token(unittest.TestCase):
    def test_each_is_unique(self):
        t1 = Filter.Token()
        t2 = Filter.Token()
        self.assertIsNot(t1, t2)
        self.assertNotEqual(t1, t2)

    def test_str(self):
        self.assertRegex(
            str(Filter.Token()),
            r"<[a-zA-Z._]+\.Filter\.Token 0x[0-9a-f]+>"
        )


class TestFilter(unittest.TestCase):
    def setUp(self):
        self.f = Filter()

    def tearDown(self):
        del self.f

    def test_register(self):
        func = unittest.mock.Mock()
        func.return_value = None

        token = self.f.register(func, 0)
        self.assertIsNotNone(token)

    def test_filter_passes_args(self):
        func = unittest.mock.Mock()
        func.return_value = None

        self.f.register(func, 0)

        iq = unittest.mock.sentinel.iq

        self.assertIsNone(self.f.filter(
            iq,
            unittest.mock.sentinel.foo,
            bar=unittest.mock.sentinel.bar,
        ))
        self.assertSequenceEqual(
            [
                unittest.mock.call(
                    iq,
                    unittest.mock.sentinel.foo,
                    bar=unittest.mock.sentinel.bar,
                ),
            ],
            func.mock_calls
        )

    def test_filter_chain(self):
        mock = unittest.mock.Mock()

        self.f.register(mock.func1, 0)
        self.f.register(mock.func2, 0)

        result = self.f.filter(
            mock.stanza,
            unittest.mock.sentinel.foo,
            unittest.mock.sentinel.bar,
            fnord=unittest.mock.sentinel.fnord,
        )

        calls = list(mock.mock_calls)

        self.assertEqual(
            mock.func2(),
            result
        )
        self.assertSequenceEqual(
            [
                unittest.mock.call.func1(
                    mock.stanza,
                    unittest.mock.sentinel.foo,
                    unittest.mock.sentinel.bar,
                    fnord=unittest.mock.sentinel.fnord,
                ),
                unittest.mock.call.func2(
                    mock.func1(),
                    unittest.mock.sentinel.foo,
                    unittest.mock.sentinel.bar,
                    fnord=unittest.mock.sentinel.fnord,
                ),
            ],
            calls
        )

    def test_filter_chain_aborts_on_None_result(self):
        mock = unittest.mock.Mock()

        mock.func2.return_value = None

        self.f.register(mock.func1, 0)
        self.f.register(mock.func2, 0)
        self.f.register(mock.func3, 0)

        result = self.f.filter(
            mock.stanza,
            unittest.mock.sentinel.foo,
            unittest.mock.sentinel.bar,
            fnord=unittest.mock.sentinel.fnord,
        )

        calls = list(mock.mock_calls)

        self.assertIsNone(result)
        self.assertSequenceEqual(
            [
                unittest.mock.call.func1(
                    mock.stanza,
                    unittest.mock.sentinel.foo,
                    unittest.mock.sentinel.bar,
                    fnord=unittest.mock.sentinel.fnord,
                ),
                unittest.mock.call.func2(
                    mock.func1(),
                    unittest.mock.sentinel.foo,
                    unittest.mock.sentinel.bar,
                    fnord=unittest.mock.sentinel.fnord,
                ),
            ],
            calls
        )

    def test_unregister_by_token(self):
        func = unittest.mock.Mock()
        token = self.f.register(func, 0)
        self.f.unregister(token)
        self.f.filter(object())
        self.assertFalse(func.mock_calls)

    def test_unregister_raises_ValueError_if_token_not_found(self):
        with self.assertRaisesRegex(ValueError, "unregistered token"):
            self.f.unregister(object())

    def test_register_with_order(self):
        mock = unittest.mock.Mock()

        self.f.register(mock.func1, 1)
        self.f.register(mock.func2, 0)
        self.f.register(mock.func3, -1)

        result = self.f.filter(mock.stanza)
        calls = list(mock.mock_calls)

        self.assertEqual(
            mock.func1(),
            result
        )
        self.assertSequenceEqual(
            [
                unittest.mock.call.func3(mock.stanza),
                unittest.mock.call.func2(mock.func3()),
                unittest.mock.call.func1(mock.func2()),
            ],
            calls
        )

    def test_context_register_is_context_manager(self):
        cm = self.f.context_register(
            unittest.mock.sentinel.func,
            unittest.mock.sentinel.order,
        )
        self.assertTrue(
            hasattr(cm, "__enter__")
        )
        self.assertTrue(
            hasattr(cm, "__exit__")
        )

    def test_context_register_enter_registers_filter(self):
        with contextlib.ExitStack() as stack:
            register = stack.enter_context(unittest.mock.patch.object(
                self.f,
                "register",
            ))
            register.return_value = unittest.mock.sentinel.token
            stack.enter_context(unittest.mock.patch.object(
                self.f,
                "unregister",
            ))

            cm = self.f.context_register(
                unittest.mock.sentinel.func,
                unittest.mock.sentinel.order
            )
            register.assert_not_called()
            cm.__enter__()

            register.assert_called_with(
                unittest.mock.sentinel.func,
                unittest.mock.sentinel.order,
            )

            cm.__exit__(None, None, None)

    def test_context_register_enter_registers_filter_without_order_if_order_not_passed(self):  # NOQA
        with contextlib.ExitStack() as stack:
            register = stack.enter_context(unittest.mock.patch.object(
                self.f,
                "register",
            ))
            register.return_value = unittest.mock.sentinel.token
            stack.enter_context(unittest.mock.patch.object(
                self.f,
                "unregister",
            ))

            cm = self.f.context_register(
                unittest.mock.sentinel.func,
            )
            register.assert_not_called()
            cm.__enter__()

            register.assert_called_with(
                unittest.mock.sentinel.func,
            )

            cm.__exit__(None, None, None)

    def test_context_register_exit_unregisters_filter(self):
        with contextlib.ExitStack() as stack:
            register = stack.enter_context(unittest.mock.patch.object(
                self.f,
                "register",
            ))
            register.return_value = unittest.mock.sentinel.token
            unregister = stack.enter_context(unittest.mock.patch.object(
                self.f,
                "unregister",
            ))

            cm = self.f.context_register(
                unittest.mock.sentinel.func,
                unittest.mock.sentinel.order,
            )

            cm.__enter__()
            unregister.assert_not_called()

            cm.__exit__(None, None, None)
            unregister.assert_called_once_with(
                unittest.mock.sentinel.token
            )

    def test_context_register_unregisters_also_on_exception(self):
        class FooException(Exception):
            pass

        with contextlib.ExitStack() as stack:
            register = stack.enter_context(unittest.mock.patch.object(
                self.f,
                "register",
            ))
            register.return_value = unittest.mock.sentinel.token
            unregister = stack.enter_context(unittest.mock.patch.object(
                self.f,
                "unregister",
            ))

            stack.enter_context(self.assertRaises(FooException))

            cm = self.f.context_register(
                unittest.mock.sentinel.func,
                unittest.mock.sentinel.order,
            )

            stack.enter_context(cm)

            raise FooException()

        unregister.assert_called_once_with(
            unittest.mock.sentinel.token
        )


class Testfirst_signal(unittest.TestCase):
    def test_connects_future_to_both_and_returns_future(self):
        s1 = unittest.mock.Mock()
        s2 = unittest.mock.Mock()

        with contextlib.ExitStack() as stack:
            Future = stack.enter_context(unittest.mock.patch("asyncio.Future"))

            result = first_signal(s1, s2)

        Future.assert_called_once_with()

        s1.connect.assert_called_once_with(
            Future(),
            s1.AUTO_FUTURE,
        )

        s2.connect.assert_called_once_with(
            Future(),
            s2.AUTO_FUTURE,
        )

        self.assertEqual(Future(), result)

    def test_works_for_common_use_case_with_success(self):
        s1 = AdHocSignal()
        s2 = AdHocSignal()

        fut = first_signal(s1, s2)

        self.assertFalse(fut.done())

        s1()

        self.assertTrue(fut.done())
        self.assertIsNone(fut.result())

    def test_works_for_common_use_case_with_exception(self):
        exc = Exception()

        s1 = AdHocSignal()
        s2 = AdHocSignal()

        fut = first_signal(s1, s2)

        self.assertFalse(fut.done())

        s2(exc)

        self.assertTrue(fut.done())
        self.assertIs(fut.exception(), exc)
